
{*****************************************}
{                                         }
{             FastReport v2.2             }
{             Report classes              }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_Class;

interface

{$I FR_Vers.inc}

// Uncomment following line, if you want storing report data in DB
//{$DEFINE StoreToDB}

uses
  SysUtils, Windows, Messages, Classes, Graphics, Printers, Controls,
  DB, Forms, StdCtrls, ComCtrls, Dialogs, Menus, Buttons,
  {$IFDEF StoreToDB} DBTables, {$ENDIF} FR_Pars, FR_DSet, FR_DBSet, FR_Utils;


const
  flStretched = 1;
  flWordWrap = 2;
  flBandNewPageAfter = 2;
  flBandPrintifSubsetEmpty = 4;
  flBandPageBreak = 8;
  flPictCenter = 2;
  flPictRatio = 4;
  gtMemo = 0;
  gtPicture = 1;
  gtBand = 2;
  gtSubReport = 3;
  gtAddIn = 10;

type
  TfrDrawMode = (drAll,drCalcHeight,drAfterCalcHeight,drPart);
  TfrBandType = (btReportTitle,btReportSummary,
               btPageHeader,btPageFooter,
               btMasterHeader,btMasterData,btMasterFooter,
               btDetailHeader,btDetailData,btDetailFooter,
               btSubDetailHeader,btSubDetailData,btSubDetailFooter,
               btOverlay,btStaticColumnHeader,btStaticColumnFooter,
               btGroupHeader,btGroupFooter,
               btColumnHeader,btColumnData,btColumnFooter,btNone);
  TfrDataSetPosition = (psLocal,psGlobal);
  TfrValueType = (vtNotAssigned,vtDBField,vtOther);
  TfrPageMode = (pmNormal,pmBuildList);
  TfrBandRecType = (rtShowBand,rtFirst,rtNext);
  TfrRgnType = (rtNormal,rtExtended);

  TfrView = class;
  TfrBand = class;
  TfrPage = class;
  TfrReport = class;
  TfrExportFilter = class;

  TDetailEvent = procedure(const ParName:String; var ParValue:Variant) of object;
  TEnterRectEvent = procedure(Memo:TStringList; View:TfrView) of object;
  TBeginDocEvent = procedure of object;
  TEndDocEvent = procedure of object;
  TBeginPageEvent = procedure(pg:Integer; MasterMemo,DetailMemo,
                              SubDetailMemo:TStringList) of object;
  TEndPageEvent = procedure(pg:Integer) of object;
  TProgressEvent = procedure(n:Integer) of object;

  TfrHighlightAttr = packed record
    FontStyle: Word;
    FontColor,FillColor: TColor;
  end;

  TfrPrnInfo = record // print info about page size, margins e.t.c
    PPgw,PPgh,Pgw,Pgh: Integer; // page width/height (printer/screen)
    POfx,POfy,Ofx,Ofy: Integer; // offset x/y
    PPw,PPh,Pw,Ph: Integer;     // printable width/height
  end;

  PfrPageInfo = ^TfrPageInfo;
  TfrPageInfo = packed record // pages of a preview
    R: TRect;
    pgSize: Word;
    pgWidth,pgHeight: Integer;
    pgOr: TPrinterOrientation;
    pgMargins: Boolean;
    PrnInfo: TfrPrnInfo;
    Pict: TMetaFile;
    Canvas: TMetafileCanvas;
  end;

  PfrBandRec = ^TfrBandRec;
  TfrBandRec = packed record
    Band: TfrBand;
    Action: TfrBandRecType;
  end;


  TfrView = class(TObject)
  private
    procedure P1Click(Sender:TObject);
  protected
    Parent: TfrBand;
    Canvas: TCanvas;
    DRect: TRect;
    gapx,gapy: Integer;
    Memo1: TStringList;
    oldx,oldy,olddy: Integer;
    procedure CalcGaps; virtual;
    procedure ShowBackGround; virtual;
    procedure ShowFrame; virtual;
    procedure BeginDraw(ACanvas:TCanvas);
  public
    Typ: Byte;
    Selected: Boolean;
    OriginalRect: TRect;
    x,y,dx,dy: Integer;
    Flags: Word;
    FrameTyp: Word;
    FrameWidth: Integer;
    FrameColor: TColor;
    FrameStyle: Word; // not used in this version
    FillColor: TColor;
    Format: Integer;
    FormatStr: String;
    Memo: TStringList;
    constructor Create; virtual;
    procedure Assign(From:TfrView); virtual;
    destructor Destroy; override;
    procedure Draw(Canvas:TCanvas); virtual; abstract;
    procedure LoadFromStream(Stream:TStream); virtual;
    procedure SaveToStream(Stream:TStream); virtual;
    procedure Resized; virtual;
    procedure GetBlob(b:TField); virtual;
    procedure DefinePopupMenu(Popup:TPopupMenu); virtual;
    function GetClipRgn(rt:TfrRgnType): HRGN; virtual;
  end;

  TfrMemoView = class(TfrView)
  private
    procedure P1Click(Sender:TObject);
    procedure P2Click(Sender:TObject);
  protected
    FUseHlt: Boolean;
    DrawMode: TfrDrawMode;
    FromLine: Integer;
    TextHeight: Integer;
    procedure ExpandVariables;
    procedure AssignFont;
    procedure WrapMemo;
    procedure ShowMemo;
  public
    FontName: String[32];
    FontSize: Integer;
    FontStyle: Word;
    FontColor: TColor;
    Adjust: Integer;
    Highlight: TfrHighlightAttr;
    HighlightStr: String;
    UseHighlight: WordBool;
    constructor Create; override;
    procedure Assign(From:TfrView); override;
    procedure Draw(Canvas:TCanvas); override;
    procedure LoadFromStream(Stream:TStream); override;
    procedure SaveToStream(Stream:TStream); override;
    procedure GetBlob(b:TField); override;
    procedure DefinePopupMenu(Popup:TPopupMenu); override;
  end;

  TfrBandView = class(TfrView)
  private
    procedure P1Click(Sender:TObject);
    procedure P2Click(Sender:TObject);
    procedure P3Click(Sender:TObject);
  public
    constructor Create; override;
    procedure Draw(Canvas:TCanvas); override;
    procedure DefinePopupMenu(Popup:TPopupMenu); override;
    function GetClipRgn(rt:TfrRgnType): HRGN; override;
  end;

  TfrSubReportView = class(TfrView)
  public
    SubPage: Integer;
    constructor Create; override;
    procedure Assign(From:TfrView); override;
    procedure Draw(Canvas:TCanvas); override;
    procedure LoadFromStream(Stream:TStream); override;
    procedure SaveToStream(Stream:TStream); override;
    procedure DefinePopupMenu(Popup:TPopupMenu); override;
  end;

  TfrPictureView = class(TfrView)
  private
    procedure P1Click(Sender:TObject);
    procedure P2Click(Sender:TObject);
  public
    Picture: TPicture;
    constructor Create; override;
    procedure Assign(From:TfrView); override;
    destructor Destroy; override;
    procedure Draw(Canvas:TCanvas); override;
    procedure LoadFromStream(Stream:TStream); override;
    procedure SaveToStream(Stream:TStream); override;
    procedure GetBlob(b:TField); override;
    procedure DefinePopupMenu(Popup:TPopupMenu); override;
  end;

  TfrBand = class(TObject)
  private
    Parent: TfrPage;
    Typ: TfrBandType;
    Next,Prev: TfrBand;
    SubIndex,MaxY: Integer;
    EOFReached: Boolean;
    EOFArr: Array[0..31] of Boolean;
    Positions: Array[TfrDatasetPosition] of Integer;
    LastGroupValue: Variant;
    LastBand,FooterBand: TfrBand;
    GroupValues: TStringList;
    Count: Integer;
    procedure InitDataSet(Desc:String);
    procedure DoError;
    function CalcHeight: Integer;
    procedure StretchObjects(MaxHeight:Integer);
    procedure UnStretchObjects;
    procedure DrawObject(t:TfrView; DrawMode:TfrDrawMode);
    procedure PrepareSubReports;
    procedure DoSubReports;
    function DrawObjects(DrawMode:TfrDrawMode): Boolean;
    procedure DrawColumnCell(Parnt:TfrBand; CurX:Integer);
    procedure DrawColumnAll;
    function CheckPageBreak(y,dy:Integer; PBreak:Boolean): Boolean;
    procedure DrawPageBreak(DrawMode:TfrDrawMode);
    function DoCalcHeight: Integer;
    procedure DoDraw;
    function Draw: Boolean;
    procedure InitGroupValues;
    procedure DoAggregate;
  public
    x,y,dx,dy,maxdy: Integer;
    PrintIfSubsetEmpty,NewPageAfter,Stretched,PageBreak,Visible: Boolean;
    Objects: TList;
    Memo: TStringList;
    DataSet: TfrDataSet;
    GroupCondition: String;
    constructor Create(ATyp:TfrBandType; AParent:TfrPage);
    destructor Destroy; override;
  end;

  TfrValue = class
  public
    Typ: TfrValueType;
    OtherKind: Integer;   // for vtOther - typ, for vtDBField - format
    DataSet: String;      // for vtDBField
    Field: String;        // here is an expression for vtOther
    DSet: TDataSet;
    DField: TField;
  end;

  TfrValues = class(TPersistent)
  private
    FItems: TStringList;
    function GetValue(Index:Integer):TfrValue;
  protected
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    function AddValue:Integer;
    function FindVariable(const s:String):TfrValue;
    procedure ReadBinaryData(Stream: TStream);
    procedure WriteBinaryData(Stream: TStream);
    procedure Clear;
    property Items: TStringList read FItems write FItems;
    property Objects[Index:Integer]:TfrValue read GetValue;
  end;

  TfrPage = class(TObject)
  private
    Bands: Array[TfrBandType] of TfrBand;
    Skip,InitFlag: Boolean;
    CurY,CurBottomY: Integer;
    CurColumn,LastStaticColumnY,XAdjust: Integer;
    List: TList;
    Mode: TfrPageMode;
    PlayFrom: Integer;
    LastBand: TfrBand;
    ColPos,CurPos: Integer;
    function TopMargin: Integer;
    function BottomMargin: Integer;
    function LeftMargin: Integer;
    function RightMargin: Integer;
    procedure InitReport;
    procedure DoneReport;
    procedure TossObjects;
    procedure FormPage;
    procedure AddRecord(b:TfrBand; rt:TfrBandRecType);
    procedure ClearRecList;
    function PlayRecList: Boolean;
    procedure ShowBand(b:TfrBand);
    procedure NewPage;
    procedure DrawPageFooters(FreeCanvases:Boolean);
    function BandExists(b:TfrBand):Boolean;
    procedure LoadFromStream(Stream:TStream);
    procedure SaveToStream(Stream:TStream);
  public
    pgSize,pgWidth,pgHeight: Integer;
    pgMargins: TRect;
    pgOr: TPrinterOrientation;
    PHonFirst,PFonLast,CHCopy,PrintToPrevPage,UseMargins: WordBool;
    PrnInfo: TfrPrnInfo;
    ColCount,ColWidth,ColGap: Integer;
    Objects: TList;
    constructor Create(ASize,AWidth,AHeight:Integer; AOr:TPrinterOrientation;
      ADoc:TfrReport);
    destructor Destroy; override;
    procedure Clear;
    procedure Delete(Index:Integer);
    procedure ChangePaper(ASize,AWidth,AHeight:Integer; AOr:TPrinterOrientation);
  end;

  TfrPages = class(TObject)
  private
    Parent: TfrReport;
    function GetCount: Integer;
    function GetPages(Index:Integer): TfrPage;
  public
    FPages: TList;
    constructor Create(AParent:TfrReport);
    destructor Destroy; override;
    procedure Clear;
    procedure Add;
    procedure Delete(Index:Integer);
    procedure LoadFromStream(Stream:TStream);
    procedure SaveToStream(Stream:TStream);
    property Pages[Index:Integer]: TfrPage read GetPages; default;
    property Count: Integer read GetCount;
  end;

  TfrEMFPages = class(TObject)
  private
    Parent: TfrReport;
    function GetCount: Integer;
    function GetPages(Index:Integer): PfrPageInfo;
  public
    FPages: TList;
    constructor Create(AParent:TfrReport);
    destructor Destroy; override;
    procedure Clear;
    procedure Add(Page:TfrPage);
    procedure LoadFromStream(Stream:TStream);
    procedure SaveToStream(Stream:TStream);
    property Pages[Index:Integer]: PfrPageInfo read GetPages; default;
    property Count: Integer read GetCount;
  end;

  TfrReport = class(TComponent)
  private
    FPages: TfrPages;
    FEMFPages: TfrEMFPages;
    FVars: TStrings;
    FVal: TfrValues;
    FTitle: String;
    FShowProgress: Boolean;
    FTemplDir: String;
    FOnBeginDoc: TBeginDocEvent;
    FOnEndDoc: TEndDocEvent;
    FOnBeginPage: TBeginPageEvent;
    FOnEndPage: TEndPageEvent;
    FOnGetValue: TDetailEvent;
    FOnEnterRect: TEnterRectEvent;
    FOnProgress: TProgressEvent;
    FOnFunction: TFunctionEvent;
    FCurrentFilter: TfrExportFilter;
    FFromPage,FToPage,Fn: Integer;
    FCollate: Boolean;
    FCurPage: TfrPage;
    function FormatValue(V:Variant; Format:Integer;
      const FormatStr:String):String;
    procedure OnGetParsFunction(const name:String; p1,p2,p3:Variant;
                                var val:String);
    procedure PrepareDataSets;
    procedure BuildBeforeModal(Sender:TObject);
    procedure PrintBeforeModal(Sender:TObject);
    procedure DoBuildReport; virtual;
    procedure DoPrintReport(FromPage,ToPage,n:Integer; Collate:Boolean);
    procedure SetPrinterTo(PrnName:String);
    procedure SetVars(Value:TStrings);
    procedure StartReport;
  protected
    procedure DefineProperties(Filer: TFiler); override;
    procedure ReadBinaryData(Stream: TStream);
    procedure WriteBinaryData(Stream: TStream);
  public
    CanRebuild: Boolean;            // true, if report can be rebuilded
    HasBlobs: Boolean;
    Terminated: Boolean;
    PrintToDefault,DoublePass: WordBool;
    FinalPass: Boolean;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
    // service methods
    function FindVariable(Variable:String):Integer;
    procedure GetVariableValue(const s:String; var v:Variant);
    procedure GetVarList(CatNo:Integer; List:TStrings);
    procedure GetCategoryList(List:TStrings);
    // internal events used through report building
    procedure InternalOnEnterRect(Memo:TStringList; View:TfrView);
    procedure InternalOnExportData(View:TfrView);
    procedure InternalOnExportText(x,y:Integer; const text:String; View:TfrView);
    procedure InternalOnGetValue(const ParName:String; var ParValue:String);
    procedure InternalOnProgress(Percent: Integer);
    // load/save methods
    procedure LoadFromStream(Stream:TStream);
    procedure SaveToStream(Stream:TStream);
    procedure LoadFromFile(FName:String);
    procedure SaveToFile(FName:String);
{$IFDEF StoreToDB}
    procedure LoadFromDB(Table:TTable; DocN:Integer);
    procedure SaveToDB(Table:TTable; DocN:Integer);
{$ENDIF}
    procedure LoadTemplate(fname:String; comm:TStrings;
      Bmp:TBitmap; Load:Boolean);
    procedure SaveTemplate(fname:String; comm:TStrings; Bmp:TBitmap);
    procedure LoadPreparedReport(FName:String);
    procedure SavePreparedReport(FName:String);
    // report manipulation methods
    procedure DesignReport;
    function PrepareReport: Boolean;
    procedure ExportTo(Filter:TClass; FileName:String);
    procedure ShowReport;
    procedure ShowPreparedReport;
    procedure PrintPreparedReport(FromPage,ToPage,n:Integer; Collate:Boolean);
    // printer manipulation methods
    function ChangePrinter(OldIndex,NewIndex:Integer):Boolean;
    //
    property Variables: TStrings read FVars write SetVars;
    property Pages: TfrPages read FPages;
    property EMFPages: TfrEMFPages read FEMFPages;
  published
    property ShowProgress: Boolean read FShowProgress write FShowProgress default True;
    property TemplateDir: String read FTemplDir write FTemplDir;
    property Title: String read FTitle write FTitle;
    property Values: TfrValues read FVal write FVal;
    property OnBeginDoc: TBeginDocEvent read FOnBeginDoc write FOnBeginDoc;
    property OnEndDoc: TEndDocEvent read FOnEndDoc write FOnEndDoc;
    property OnBeginPage: TBeginPageEvent read FOnBeginPage write FOnBeginPage;
    property OnEndPage: TEndPageEvent read FOnEndPage write FOnEndPage;
    property OnGetValue: TDetailEvent read FOnGetValue write FOnGetValue;
    property OnEnterRect: TEnterRectEvent read FOnEnterRect write FOnEnterRect;
    property OnUserFunction: TFunctionEvent read FOnFunction write FOnFunction;
    property OnProgress: TProgressEvent read FOnProgress write FOnProgress;
  end;

  TfrCompositeReport = class(TfrReport)
  private
    procedure DoBuildReport; override;
  public
    Reports: TList;
    constructor Create(AOwner:TComponent); override;
    destructor Destroy; override;
  end;

  TfrReportDesigner = class(TSForm)
  public
    Page: TfrPage;
    procedure RegisterObject(ButtonBmp:TBitmap; const ButtonHint:String;
      ButtonTag:Integer); virtual; abstract;
    procedure PopupNotify(Sender: TfrView); virtual; abstract;
  end;

  TfrObjEditorForm = class(TSForm)
  public
    procedure ShowEditor(t:TfrView); virtual;
  end;

  TfrExportFilter = class(TObject)
  protected
    Stream: TStream;
    DataRow,TextRow,LastDataRow,LastTextRow: Integer;
    RowTop,FontHeight: Integer;
  public
    constructor Create(AStream:TStream); virtual;
    procedure OnBeginDoc; virtual;
    procedure OnEndDoc; virtual;
    procedure OnBeginPage; virtual;
    procedure OnEndPage; virtual;
    procedure OnData(x,y:Integer; View:TfrView); virtual;
    procedure OnText(x,y:Integer; const text:String; View:TfrView); virtual;
  end;

  TfrFunctionLibrary = class(TObject)
  public
    List: TStringList;
    constructor Create; virtual;
    destructor Destroy; override;
    function OnFunction(const FName:String; p1,p2,p3:Variant;
      var val:String): Boolean;
    procedure DoFunction(FNo:Integer; p1,p2,p3:Variant; var val:String);
      virtual; abstract;
  end;


function frCreateObject(Typ:Byte; const ClassName:String):TfrView;
procedure frRegisterObject(ClassRef:TClass; ButtonBmp:TBitmap;
  const ButtonHint:String; EditorForm:TfrObjEditorForm);
procedure frRegisterExportFilter(ClassRef:TClass;
  const FilterDesc,FilterExt: String);
procedure frRegisterFunctionLibrary(ClassRef:TClass);


const
  frCurrentVersion = 22; // this is current version (2.2)
  frSpecCount = 9;
  frSpecFuncs: Array[0..frSpecCount-1] of String = ('PAGE#','',
    'DATE','TIME','LINE#','LINETHROUGH#','COLUMN#','CURRENT#','TOTALPAGES');

type
  TfrAddInObjectInfo = record
    ClassRef: TClass;
    EditorForm: TfrObjEditorForm;
    ButtonBmp: TBitmap;
    ButtonHint: String;
  end;

  TfrExportFilterInfo = record
    ClassRef: TClass;
    FilterDesc,FilterExt: String;
  end;

  TfrFunctionInfo = record
    FunctionLibrary: TfrFunctionLibrary;
  end;

var
  frDesigner: TfrReportDesigner;                  // designer reference
  CurReport: TfrReport;                           // currently proceeded report
  DocMode: (dmDesigning, dmPrinting);             // current mode
  DisableDrawing: Boolean;
  frAddIns: Array[0..31] of TfrAddInObjectInfo;   // add-in objects
  frAddInsCount: Integer;
  frFilters: Array[0..31] of TfrExportFilterInfo; // export filters
  frFiltersCount: Integer;
  frFunctions: Array[0..31] of TfrFunctionInfo;   // function libraries
  frFunctionsCount: Integer;
  PageNo: Integer;                  // current page number in Building mode
  frCharset: 0..255;
  frBandNames: Array[0..21] of String;
  frSpecArr: Array[0..frSpecCount-1] of String;
  frVersion: Byte;                  // version of currently loaded report

implementation

uses FR_Fmted, FR_View, FR_Prntr, FR_Progr, FR_EText;

const
  pkNone = 0;
  pkBitmap = 1;
  pkMetafile = 2;
  pkIcon = 3;

type
  TfrBandParts = (bpHeader,bpData,bpFooter);

  TfrStdFunctionLibrary = class(TfrFunctionLibrary)
  public
    constructor Create; override;
    procedure DoFunction(FNo:Integer; p1,p2,p3:Variant; var val:String); override;
  end;

var
  SHeight,VHeight: Integer;   // used for height calculation of TfrMemoView
  SMemo: TStringList;         // temporary memo used during TfrView drawing
  SBmp: TBitmap;              // small bitmap used by TfrBandView drawing
  CurDate,CurTime: TDateTime; // date/time of report starting
  CurView: TfrView;           // currently proceeded view
  CurBand: TfrBand;           // currently proceeded band
  CurPage: TfrPage;           // currently proceeded page
  CurValue: Variant;          // used for highlighting
  CurVariable: String;
  CurDataRow,CurStrNo,CurFontHeight: Integer; // used for export
  SavedAllPages: Integer;     // number of pages in entire report
  MasterReport: TfrReport;    // reference to main composite report
  ErrorFlag: Boolean;         // error occured through TfrView drawing
  ErrorStr: String;           // error description
  SubValue: String;           // used in GetValue event handler

  // variables used through report building
  MCanvas: TMetafileCanvas;
  PrevY,PrevBottomY,ColumnXAdjust: Integer;
  Append,WasPF: Boolean;
  CompositeMode: Boolean;

{----------------------------------------------------------------------------}
function frCreateObject(Typ:Byte; const ClassName:String): TfrView;
var i: Integer;
begin
  Result := nil;
  case Typ of
    gtMemo:      Result := TfrMemoView.Create;
    gtPicture:   Result := TfrPictureView.Create;
    gtBand:      Result := TfrBandView.Create;
    gtSubReport: Result := TfrSubReportView.Create;
    gtAddIn:
      begin
        for i := 0 to frAddInsCount-1 do
          if frAddIns[i].ClassRef.ClassName = ClassName then
          begin
            Result := TfrView(frAddIns[i].ClassRef.NewInstance);
            Result.Create;
            Result.Typ := gtAddIn;
            break;
          end;
        if Result = nil then
          raise EClassNotFound.Create('   ' + ClassName);
      end;
  end;
end;

procedure frRegisterObject(ClassRef:TClass; ButtonBmp:TBitmap;
  const ButtonHint:String; EditorForm:TfrObjEditorForm);
begin
  frAddIns[frAddInsCount].ClassRef := ClassRef;
  frAddIns[frAddInsCount].EditorForm := EditorForm;
  frAddIns[frAddInsCount].ButtonBmp := ButtonBmp;
  frAddIns[frAddInsCount].ButtonHint := ButtonHint;
  if frDesigner <> nil then
    frDesigner.RegisterObject(ButtonBmp, ButtonHint,
      Integer(gtAddIn) + frAddInsCount);
  Inc(frAddInsCount);
end;

procedure frRegisterExportFilter(ClassRef:TClass;
  const FilterDesc,FilterExt:String);
begin
  frFilters[frFiltersCount].ClassRef := ClassRef;
  frFilters[frFiltersCount].FilterDesc := FilterDesc;
  frFilters[frFiltersCount].FilterExt := FilterExt;
  Inc(frFiltersCount);
end;

procedure frRegisterFunctionLibrary(ClassRef:TClass);
begin
  frFunctions[frFunctionsCount].FunctionLibrary :=
    TfrFunctionLibrary(ClassRef.NewInstance);
  frFunctions[frFunctionsCount].FunctionLibrary.Create;
  Inc(frFunctionsCount);
end;

function Create90Font(Font:TFont):HFont;
var F: TLogFont;
begin
  GetObject(Font.Handle, SizeOf(TLogFont), @F);
  F.lfEscapement := 900;
  F.lfOrientation := 900;
  Result := CreateFontIndirect(F);
end;

function GetDefaultDataSet: TDataSet;
var Res: TfrDataset;
begin
  Result := nil; Res := nil;
  case CurBand.Typ of
    btMasterData, btReportSummary, btMasterFooter,
    btGroupHeader, btGroupFooter:
      Res := CurPage.Bands[btMasterData].DataSet;
    btDetailData, btDetailFooter:
      Res := CurPage.Bands[btDetailData].DataSet;
    btSubDetailData, btSubDetailFooter:
      Res := CurPage.Bands[btSubDetailData].DataSet;
    btColumnData, btColumnFooter:
      Res := CurPage.Bands[btColumnData].DataSet;
  end;
  if (Res <> nil) and (Res is TfrDBDataset) then
    with Res as TfrDBDataset do
      if DataSource <> nil then
        Result := DataSource.DataSet;
end;

{----------------------------------------------------------------------------}
constructor TfrView.Create;
begin
  inherited Create;
  Parent := nil;
  Memo := TStringList.Create;
  Memo1 := TStringList.Create;
  FrameColor := clBlack;
  FillColor := clNone;
  Format := 2*256 + Ord(DecimalSeparator);
end;

procedure TfrView.Assign(From:TfrView);
begin
  Typ := From.Typ;
  Selected := From.Selected;
  x := From.x; y := From.y; dx := From.dx; dy := From.dy;
  Flags := From.Flags;
  FrameTyp := From.FrameTyp;
  FrameWidth := From.FrameWidth;
  FrameColor := From.FrameColor;
  FrameStyle := From.FrameStyle;
  FillColor := From.FillColor;
  Format := From.Format;
  FormatStr := From.FormatStr;
  Memo.Assign(From.Memo);
end;

destructor TfrView.Destroy;
begin
  Memo.Free;
  Memo1.Free;
  inherited Destroy;
end;

procedure TfrView.CalcGaps;
var bx,by,bx1,by1,wx1,wx2: Integer;
begin
  wx1 := (FrameWidth-1) div 2; wx2 := FrameWidth div 2;
  gapx := wx2 + 2; gapy := wx2 div 2 + 1;
  bx := x; by := y; bx1 := x+dx; by1 := y+dy;
  if (FrameTyp and $1) <> 0 then Dec(bx1, wx2);
  if (FrameTyp and $2) <> 0 then Dec(by1, wx2);
  if (FrameTyp and $4) <> 0 then Inc(bx, wx1);
  if (FrameTyp and $8) <> 0 then Inc(by, wx1);
  DRect := Rect(bx, by, bx1+1, by1+1);
end;

procedure TfrView.ShowBackground;
var fp: TColor;
begin
  if DisableDrawing then Exit;
  if (DocMode = dmPrinting) and (FillColor = clNone) then Exit;
  fp := FillColor;
  if (DocMode = dmDesigning) and (fp = clNone) then
    fp := clWhite;
  Canvas.Brush.Color := fp;
  if DocMode = dmDesigning then
    Canvas.FillRect(DRect) else
    Canvas.FillRect(Rect(x, y, x+dx, y+dy));
end;

procedure TfrView.ShowFrame;
procedure Line(x,y,dx,dy:Integer);
begin
  Canvas.MoveTo(x, y);
  Canvas.LineTo(x+dx, y+dy);
end;
begin
  if DisableDrawing then Exit;
  if (DocMode = dmPrinting) and ((FrameTyp and $F) = 0) then Exit;
  with Canvas do
  begin
    Pen.Width := FrameWidth;
    Pen.Color := FrameColor;
    Pen.Style := psSolid;
    Brush.Style := bsClear;
    if ((FrameTyp and $F) = 0) and (DocMode = dmDesigning) then
    begin
      Pen.Color := clBlack;
      Pen.Width := 1;
      Line(x, y+3, 0, -3); Line(x, y, 4, 0);
      Line(x, y+dy-4, 0, 3); Line(x, y+dy-1, 4, 0);
      Line(x+dx-4, y, 3, 0); Line(x+dx-1, y, 0, 4);
      Line(x+dx-4, y+dy-1, 3, 0); Line(x+dx-1, y+dy-1, 0, -4);
    end
    else
      if (FrameTyp and $F) = $F then Rectangle(x, y, x+dx+1, y+dy+1)
      else
      begin
        if (FrameTyp and $1) <> 0 then Line(x+dx, y, 0, dy);
        if (FrameTyp and $4) <> 0 then Line(x, y, 0, dy);
        if (FrameTyp and $2) <> 0 then Line(x, y+dy, dx, 0);
        if (FrameTyp and $8) <> 0 then Line(x, y, dx, 0);
      end;
  end;
end;

procedure TfrView.BeginDraw(ACanvas:TCanvas);
begin
  Canvas := ACanvas;
  CurView := Self;
end;

procedure TfrView.LoadFromStream(Stream:TStream);
begin
  with Stream do
  begin
    Read(x, 4); Read(y, 4); Read(dx, 4); Read(dy, 4);
    Read(Flags, 2);
    Read(FrameTyp, 2);
    Read(FrameWidth, 4);
    Read(FrameColor, 4);
    Read(FrameStyle, 2);
    Read(FillColor, 4);
    Read(Format, 4);
    FormatStr := '';
    frReadMemo(Stream, SMemo);
    if SMemo.Count > 0 then
      FormatStr := SMemo[0];
    frReadMemo(Stream, Memo);
  end;
end;

procedure TfrView.SaveToStream(Stream:TStream);
begin
  with Stream do
  begin
    Write(x, 4); Write(y, 4); Write(dx, 4); Write(dy, 4);
    Write(Flags, 2);
    Write(FrameTyp, 2);
    Write(FrameWidth, 4);
    Write(FrameColor, 4);
    Write(FrameStyle, 2);
    Write(FillColor, 4);
    Write(Format, 4);
    SMemo.Clear; SMemo.Add(FormatStr);
    frWriteMemo(Stream, SMemo);
    frWriteMemo(Stream, Memo);
  end;
end;

procedure TfrView.Resized;
begin
end;

procedure TfrView.GetBlob(b:TField);
begin
end;

function TfrView.GetClipRgn(rt:TfrRgnType): HRGN;
var bx,by,bx1,by1,w1,w2: Integer;
begin
  w1 := FrameWidth shr 1; w2 := (FrameWidth - 1) shr 1;
  bx := x; by := y; bx1 := x + dx + 1; by1 := y + dy + 1;
  if (FrameTyp and $1) <> 0 then Inc(bx1, w2);
  if (FrameTyp and $2) <> 0 then Inc(by1, w2);
  if (FrameTyp and $4) <> 0 then Dec(bx, w1);
  if (FrameTyp and $8) <> 0 then Dec(by, w1);
  if rt = rtNormal then
    Result := CreateRectRgn(bx, by, bx1, by1) else
    Result := CreateRectRgn(bx - 10, by - 10, bx1 + 10, by1 + 10);
end;

procedure TfrView.DefinePopupMenu(Popup:TPopupMenu);
var m: TMenuItem;
begin
  m := TMenuItem.Create(Popup);
  m.Caption := '-';
  Popup.Items.Add(m);

  m := TMenuItem.Create(Popup);
  m.Caption := frLoadStr('SStretched');
  m.OnClick := P1Click;
  m.Checked := (Flags and flStretched) <> 0;
  Popup.Items.Add(m);
end;

procedure TfrView.P1Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flStretched) + Word(Checked);
    end;
  end;
  frDesigner.PopupNotify(Self);
end;

{----------------------------------------------------------------------------}
constructor TfrMemoView.Create;
begin
  inherited Create;
  Typ := gtMemo;
  FontColor := clBlack;
  Highlight.FontColor := clBlack;
  Highlight.FillColor := clWhite;
  Highlight.FontStyle := 2; // fsBold
end;

procedure TfrMemoView.Assign(From:TfrView);
begin
  inherited Assign(From);
  FontName := (From as TfrMemoView).FontName;
  FontSize := (From as TfrMemoView).FontSize;
  FontStyle := (From as TfrMemoView).FontStyle;
  FontColor := (From as TfrMemoView).FontColor;
  Adjust := (From as TfrMemoView).Adjust;
  Highlight := (From as TfrMemoView).Highlight;
  HighlightStr := (From as TfrMemoView).HighlightStr;
  UseHighlight := (From as TfrMemoView).UseHighlight;
end;

procedure TfrMemoView.ExpandVariables;
var i: Integer;
    s: String;
procedure GetData(var s:String);
var i,j: Integer;
    s1,s2: String;
begin
  i := 1;
  repeat
    while (i < Length(s)) and (s[i] <> '[') do Inc(i);
    s1 := GetBrackedVariable(s, i, j);
    if i <> j then
    begin
      Delete(s, i, j-i+1);
      s2 := '';
      CurReport.InternalOnGetValue(s1, s2);
      Insert(s2, s, i);
      Inc(i, Length(s2));
      j := 0;
    end;
  until i = j;
end;
begin
  Memo1.Clear;
  for i := 0 to Memo.Count-1 do
  begin
    s := Memo[i];
    if DocMode <> dmDesigning then GetData(s);
    Memo1.Add(s);
  end;
end;

procedure TfrMemoView.AssignFont;
var FSt: TFontStyles;
begin
  with Canvas do
  begin
    Brush.Style := bsClear;
    Font.Name := FontName;
    Font.Size := FontSize;
    Font.Color := FontColor;
    FSt := [];
    if (FontStyle and $1) <> 0 then FSt := FSt + [fsItalic];
    if (FontStyle and $2) <> 0 then FSt := FSt + [fsBold];
    if (FontStyle and $4) <> 0 then FSt := FSt + [fsUnderLine];
    Font.Style := FSt;
  end;
end;

procedure TfrMemoView.WrapMemo;
var size,size1,maxwidth: Integer;
  procedure OutLine(const str:String);
  begin
    SMemo.Add(str);
    Inc(size, size1);
  end;
  procedure WrapLine(const s:String);
  var cur,beg,last: Integer;
  begin
    last := 1; beg := 1;
    if (Length(s) <= 1) or (Canvas.TextWidth(s) <= maxwidth) then
      OutLine(s)
    else
    begin
      for cur := 1 to Length(s) do
      begin
        if Canvas.TextWidth(Copy(s, beg, cur-beg+1)) >= maxwidth then
        begin
          if last = beg then last := cur;
          OutLine(Copy(s, beg, last-beg+1));
          if last = cur then
          begin
            beg := cur;
            break;
          end;
          beg := last+1; last := beg;
        end;
        if s[cur] in [' ','.',',','-'] then last := cur;
      end;
      if beg <> cur then OutLine(Copy(s, beg, cur-beg+1));
    end;
  end;
  procedure OutMemo;
  var i: Integer;
  begin
    size := y + gapy;
    size1 := -Canvas.Font.Height + 2;
    maxwidth := dx - gapx - gapx;
    for i := 0 to Memo1.Count-1 do
      if (Flags and flWordWrap) <> 0 then
        WrapLine(Memo1[i]) else
        OutLine(Memo1[i]);
    VHeight := size - (y+gapy) + gapy + gapy;
    TextHeight := size1;
  end;
  procedure OutMemo90;
  var i: Integer;
      h,oldh: HFont;
  begin
    h := Create90Font(Canvas.Font);
    oldh := SelectObject(Canvas.Handle,h);
    size := x + gapx;
    size1 := -Canvas.Font.Height + 2;
    maxwidth := dy - gapy - gapy;
    for i := 0 to Memo1.Count-1 do
      if (Flags and flWordWrap) <> 0 then
        WrapLine(Memo1[i]) else
        OutLine(Memo1[i]);
    SelectObject(Canvas.Handle,oldh);
    DeleteObject(h);
    VHeight := size - (x+gapx) + gapx + gapx;
  end;
begin
  AssignFont;
  SMemo.Clear;
  if (Adjust and $4) <> 0 then OutMemo90 else OutMemo;
  Memo1.Assign(SMemo);
end;

procedure TfrMemoView.ShowMemo;
var DR: TRect;
    ad,ox,oy: Integer;
procedure OutMemo;
var i,cury,th: Integer;
  function OutLine(const str:String): Boolean;
  var curx: Integer;
  begin
    if cury + th <= DR.Bottom then
    begin
      if Adjust = 0 then
        curx := x + gapx
      else if Adjust = 1 then
        curx := x + dx - 1 - gapx - Canvas.TextWidth(str)
      else
        curx := x + gapx + (dx - gapx - gapx - Canvas.TextWidth(str)) div 2;
      if (Flags and flWordWrap) <> 0 then
        ExtTextOut(Canvas.Handle, curx, cury, ETO_CLIPPED, @DR,
          PChar(str), Length(str), nil)
      else
        TextOut(Canvas.Handle, curx, cury, PChar(str), Length(str));
      CurFontHeight := -Canvas.Font.Height;
      CurReport.InternalOnExportText(curx, cury, str, Self);
      Inc(CurStrNo);
      Result := False;
    end
    else Result := True;
    cury := cury + th;
  end;
begin
  cury := y + gapy;
  th := -Canvas.Font.Height + 2;
  CurStrNo := 0;
  for i := FromLine to Memo1.Count-1 do
    if OutLine(Memo1[i]) then
    begin
      FromLine := i;
      break;
    end;
end;

procedure OutMemo90;
var i,th,curx: Integer;
    h,oldh: HFont;
  procedure OutLine(const str:String);
  var cury: Integer;
  begin
    if Adjust = 4 then
      cury := y + dy - gapy
    else if Adjust = 5 then
      cury := y + gapy + Canvas.TextWidth(str)
    else
      cury := y + dy - 1 - gapy - (dy - gapy - gapy - Canvas.TextWidth(str)) div 2;
    if (Flags and flWordWrap) <> 0 then
      ExtTextOut(Canvas.Handle, curx, cury, ETO_CLIPPED, @DR,
        PChar(str), Length(str), nil)
    else
      TextOut(Canvas.Handle, curx, cury, PChar(str), Length(str));
    CurFontHeight := -Canvas.Font.Height;
    CurReport.InternalOnExportText(curx, cury, str, Self);
    Inc(CurStrNo);
    curx := curx + th;
  end;
begin
  h := Create90Font(Canvas.Font);
  oldh := SelectObject(Canvas.Handle,h);
  curx := x + gapx;
  th := -Canvas.Font.Height + 2;
  CurStrNo := 0;
  for i := 0 to Memo1.Count-1 do
    OutLine(Memo1[i]);
  SelectObject(Canvas.Handle, oldh);
  DeleteObject(h);
end;

begin
  AssignFont;
  DR := Rect(DRect.Left+1, DRect.Top, DRect.Right-2, DRect.Bottom-1);
  if (Adjust and $18) <> 0 then
  begin
    ad := Adjust;
    ox := x; oy := y;
    Adjust := Adjust and $7;
    if (ad and $4) <> 0 then
    begin
      if (ad and $18) = $8 then
        x := x + (dx - VHeight) div 2
      else if (ad and $18) = $10 then
        x := x + dx - VHeight;
      OutMemo90;
    end
    else
    begin
      if (ad and $18) = $8 then
        y := y + (dy - VHeight) div 2
      else if (ad and $18) = $10 then
        y := y + dy - VHeight;
      OutMemo;
    end;
    Adjust := ad;
    x := ox; y := oy;
  end
  else if (Adjust and $4) <> 0 then OutMemo90 else OutMemo;
end;

procedure TfrMemoView.Draw(Canvas:TCanvas);
var s: String;
    CanExpandVar: Boolean;
    OldFontStyle: Word;
    OldFontColor,OldFillColor: TColor;
    OldFlags: Word;
begin
  BeginDraw(Canvas);
  OldFlags := Flags;
  if DocMode = dmDesigning then
    Flags := Flags or flWordWrap;
  CanExpandVar := True;
  if (DrawMode <> drPart) or (DocMode = dmDesigning) then FromLine := 0;
  if DrawMode in [drAll, drCalcHeight] then
    FUseHlt := UseHighlight;
  if DocMode = dmPrinting then
  begin
    if DrawMode in [drAll, drCalcHeight] then
    begin
      if Assigned(CurReport.OnEnterRect) or CurReport.HasBlobs then
      begin
        s := '';
        Memo1.Assign(Memo);
        if Memo1.Count > 0 then s := Memo1[0];
        CurReport.InternalOnEnterRect(Memo1, Self);
        if (Memo1.Count > 0) and (s <> Memo1[0]) then CanExpandVar := False;
      end;
    end
    else if DrawMode in [drAfterCalcHeight, drPart] then
      CanExpandVar := False;
  end;
  if CanExpandVar then ExpandVariables;
  OldFontStyle := 0;
  OldFontColor := 0;
  OldFillColor := 0;
  if (DocMode = dmPrinting) and UseHighlight and
     (DrawMode in [drAll, drAfterCalcHeight, drPart]) then
  begin
    OldFontStyle := FontStyle; FontStyle := Highlight.FontStyle;
    OldFontColor := FontColor; FontColor := Highlight.FontColor;
    OldFillColor := FillColor; FillColor := Highlight.FillColor;
  end;
  CalcGaps;
  with Canvas do
  begin
    if DrawMode <> drCalcHeight then ShowBackground;
    if Memo1.Count <> 0 then
    begin
      if DrawMode in [drAll, drCalcHeight] then WrapMemo;
      if (DrawMode <> drCalcHeight) and not DisableDrawing then
        ShowMemo;
    end;
    if DrawMode <> drCalcHeight then ShowFrame;
  end;
  if DrawMode = drCalcHeight then
    if y + VHeight + 2 > SHeight then SHeight := y + VHeight + 2;
  if (DrawMode in [drAll, drAfterCalcHeight, drPart]) and (DocMode = dmPrinting) then
  begin
    CurReport.InternalOnExportData(Self);
    if UseHighlight then
    begin
      FontStyle := OldFontStyle;
      FontColor := OldFontColor;
      FillColor := OldFillColor;
    end;
    UseHighlight := FUseHlt;
  end;
  Flags := OldFlags;
end;

procedure TfrMemoView.LoadFromStream(Stream:TStream);
begin
  inherited LoadFromStream(Stream);
  frReadMemo(Stream, SMemo);
  FontName := SMemo[0];
  with Stream do
  begin
    Read(FontSize, 4);
    Read(FontStyle, 2);
    Read(FontColor, 4);
    Read(Adjust, 4);
    Read(Highlight, 10);
    Read(UseHighlight, 2);
    frReadMemo(Stream, SMemo);
    if SMemo.Count > 0 then
      HighlightStr := SMemo[0] else
      HighlightStr := '';
  end;
  if frVersion = 21 then
    Flags := Flags or flWordWrap;
end;

procedure TfrMemoView.SaveToStream(Stream:TStream);
begin
  inherited SaveToStream(Stream);
  SMemo.Clear; SMemo.Add(FontName);
  frWriteMemo(Stream, SMemo);
  with Stream do
  begin
    Write(FontSize, 4);
    Write(FontStyle, 2);
    Write(FontColor, 4);
    Write(Adjust, 4);
    Write(Highlight, 10);
    Write(UseHighlight, 2);
    SMemo.Clear; SMemo.Add(HighlightStr);
    frWriteMemo(Stream, SMemo);
  end;
end;

procedure TfrMemoView.GetBlob(b:TField);
begin
  Memo1.Assign(b);
end;

procedure TfrMemoView.DefinePopupMenu(Popup:TPopupMenu);
var m: TMenuItem;
begin
  m := TMenuItem.Create(Popup);
  m.Caption := frLoadStr('SVarFormat');
  m.OnClick := P1Click;
  Popup.Items.Add(m);
  inherited DefinePopupMenu(Popup);

  m := TMenuItem.Create(Popup);
  m.Caption := frLoadStr('SWordWrap');
  m.OnClick := P2Click;
  m.Checked := (Flags and flWordWrap) <> 0;
  Popup.Items.Add(m);
end;

procedure TfrMemoView.P1Click(Sender:TObject);
var t: TfrView;
    i: Integer;
begin
  FmtForm := TFmtForm.Create(nil);
  t := Self;
  with FmtForm do
  begin
    Format := t.Format;
    CB1.Checked := (t as TfrMemoView).UseHighlight;
    Edit1.Text := t.FormatStr;
    Edit2.Text := (t as TfrMemoView).HighlightStr;
    if ShowModal = mrOk then
      for i := 0 to frDesigner.Page.Objects.Count-1 do
      begin
        t := frDesigner.Page.Objects[i];
        if t.Selected then
        begin
          (t as TfrMemoView).Format := Format;
          (t as TfrMemoView).FormatStr := Edit1.Text;
          (t as TfrMemoView).HighlightStr := Edit2.Text;
          (t as TfrMemoView).UseHighlight := CB1.Checked;
        end;
      end;
  end;
  FmtForm.Free;
end;

procedure TfrMemoView.P2Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flWordWrap) + Word(Checked) * flWordWrap;
    end;
  end;
end;

{----------------------------------------------------------------------------}
constructor TfrBandView.Create;
begin
  inherited Create;
  Typ := gtBand;
  Format := 0;
end;

procedure TfrBandView.Draw(Canvas:TCanvas);
var h,oldh: HFont;
begin
  FrameWidth := 1;
  if TfrBandType(FrameTyp) in [btColumnHeader..btColumnFooter] then
  begin
    y := 0; dy := frDesigner.Page.PrnInfo.Pgh;
  end
  else
  begin
    x := 0; dx := frDesigner.Page.PrnInfo.Pgw;
  end;
  BeginDraw(Canvas);
  CalcGaps;
  with Canvas do
  begin
    Brush.Bitmap := SBmp;
    FillRect(DRect);
    Font.Name := 'Arial';
    Font.Style := [];
    Font.Size := 8;
    Font.Color := clBlack;
    Pen.Width := 1;
    Pen.Color := clBtnFace;
    Pen.Style := psSolid;
    Brush.Style := bsClear;
    Rectangle(x, y, x+dx+1, y+dy+1);
    Brush.Color := clBtnFace;
    if TfrBandType(FrameTyp) in [btColumnHeader..btColumnFooter] then
    begin
      FillRect(Rect(x-18, y, x, y+100));
      Pen.Color := clBtnShadow;
      MoveTo(x-18, y+98); LineTo(x, y+98);
      Pen.Color := clBlack;
      MoveTo(x-18, y+99); LineTo(x, y+99);
      Pen.Color := clBtnHighlight;
      MoveTo(x-18, y+99); LineTo(x-18, y);
      h := Create90Font(Font);
      oldh := SelectObject(Handle, h);
      TextOut(x-15, y+TextWidth(frBandNames[FrameTyp])+4, frBandNames[FrameTyp]);
      SelectObject(Handle, oldh);
      DeleteObject(h);
    end
    else
    begin
      FillRect(Rect(x, y-18, x+100, y));
      Pen.Color := clBtnShadow;
      MoveTo(x+98, y-18); LineTo(x+98, y);
      Pen.Color := clBlack;
      MoveTo(x+99, y-18); LineTo(x+99, y);
      TextOut(x+4, y-17, frBandNames[FrameTyp]);
    end;
  end;
end;

function TfrBandView.GetClipRgn(rt:TfrRgnType): HRGN;
var R: HRGN;
begin
  if rt = rtNormal then
    Result := CreateRectRgn(x, y, x+dx+1, y+dy+1) else
    Result := CreateRectRgn(x-10, y-10, x+dx+10, y+dy+10);
  if TfrBandType(FrameTyp) in [btColumnHeader..btColumnFooter] then
    R := CreateRectRgn(x-18, y, x, y+100)
  else
    R := CreateRectRgn(x, y-18, x+100, y);
  CombineRgn(Result, Result, R, RGN_OR);
  DeleteObject(R);
end;

procedure TfrBandView.DefinePopupMenu(Popup:TPopupMenu);
var m: TMenuItem;
    b: TfrBandType;
begin
  b := TfrBandType(FrameTyp);
  if b in [btReportTitle, btReportSummary, btPageHeader, btColumnHeader,
    btMasterHeader..btSubDetailFooter, btGroupHeader, btGroupFooter] then
    inherited DefinePopupMenu(Popup);

  if b in [btReportTitle, btReportSummary, btMasterData, btDetailData,
    btSubDetailData, btMasterFooter, btDetailFooter,
    btSubDetailFooter, btGroupHeader] then
  begin
    m := TMenuItem.Create(Popup);
    m.Caption := frLoadStr('SFormNewPage');
    m.OnClick := P1Click;
    m.Checked := (Flags and flBandNewPageAfter) <> 0;
    Popup.Items.Add(m);
  end;

  if b in [btMasterData, btDetailData] then
  begin
    m := TMenuItem.Create(Popup);
    m.Caption := frLoadStr('SPrintIfSubsetEmpty');
    m.OnClick := P2Click;
    m.Checked := (Flags and flBandPrintIfSubsetEmpty) <> 0;
    Popup.Items.Add(m);
  end;

  if b in [btReportTitle, btReportSummary, btMasterHeader..btSubDetailFooter,
    btGroupHeader, btGroupFooter] then
  begin
    m := TMenuItem.Create(Popup);
    m.Caption := frLoadStr('SBreaked');
    m.OnClick := P3Click;
    m.Checked := (Flags and flBandPageBreak) <> 0;
    Popup.Items.Add(m);
  end;
end;

procedure TfrBandView.P1Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flBandNewPageAfter) +
          Word(Checked) * flBandNewPageAfter;
    end;
  end;
end;

procedure TfrBandView.P2Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flBandPrintifSubsetEmpty) +
          Word(Checked) * flBandPrintifSubsetEmpty;
    end;
  end;
end;

procedure TfrBandView.P3Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flBandPageBreak) + Word(Checked) * flBandPageBreak;
    end;
  end;
end;

{----------------------------------------------------------------------------}
constructor TfrSubReportView.Create;
begin
  inherited Create;
  Typ := gtSubReport;
end;

procedure TfrSubReportView.Assign(From:TfrView);
begin
  inherited Assign(From);
  SubPage := (From as TfrSubReportView).SubPage;
end;

procedure TfrSubReportView.Draw(Canvas:TCanvas);
begin
  BeginDraw(Canvas);
  FrameWidth := 1;
  CalcGaps;
  with Canvas do
  begin
    Font.Name := 'Arial';
    Font.Style := [];
    Font.Size := 8;
    Font.Color := clBlack;
    Pen.Width := 1;
    Pen.Color := clBlack;
    Pen.Style := psSolid;
    Brush.Color := clWhite;
    Rectangle(x, y, x+dx+1, y+dy+1);
    Brush.Style := bsClear;
    TextRect(DRect, x+2, y+2, frLoadStr('SSubReportOnPage') + ' ' + IntToStr(SubPage+1));
  end;
end;

procedure TfrSubReportView.DefinePopupMenu(Popup:TPopupMenu);
begin
  // no specific items in popup
end;

procedure TfrSubReportView.LoadFromStream(Stream:TStream);
begin
  inherited LoadFromStream(Stream);
  Stream.Read(SubPage, 4);
end;

procedure TfrSubReportView.SaveToStream(Stream:TStream);
begin
  inherited SaveToStream(Stream);
  Stream.Write(SubPage, 4);
end;

{----------------------------------------------------------------------------}
constructor TfrPictureView.Create;
begin
  inherited Create;
  Typ := gtPicture;
  Picture := TPicture.Create;
  Flags := flStretched + flPictRatio;
end;

procedure TfrPictureView.Assign(From:TfrView);
begin
  inherited Assign(From);
  Picture.Assign((From as TfrPictureView).Picture);
end;

destructor TfrPictureView.Destroy;
begin
  Picture.Free;
  inherited Destroy;
end;

procedure TfrPictureView.Draw(Canvas:TCanvas);
var r: TRect;
    kx,ky: Double;
    w,h,w1,h1: Integer;
begin
  BeginDraw(Canvas);
  SMemo.Assign(Memo);
  if DocMode = dmPrinting then
    if Assigned(CurReport.OnEnterRect) or CurReport.HasBlobs then
      CurReport.InternalOnEnterRect(SMemo, Self);
  CalcGaps;
  with Canvas do
  begin
    ShowBackground;
    if (DocMode = dmDesigning) and
     ((Picture.Graphic = nil) or Picture.Graphic.Empty) then
    begin
      Font.Name := 'Arial';
      Font.Size := 8;
      Font.Style := [];
      Font.Color := clBlack;
      TextOut(x+2, y+2, frLoadStr('SPicture'));
    end
    else if not ((Picture.Graphic = nil) or Picture.Graphic.Empty) then
    begin
      if (Flags and flStretched) <> 0 then
      begin
        r := DRect;
        if (Flags and flPictRatio) <> 0 then
        begin
          kx := dx / Picture.Width;
          ky := dy / Picture.Height;
          if kx < ky then
            r := Rect(DRect.Left, DRect.Top,
              DRect.Right, DRect.Top + Round(Picture.Height * kx))
          else
            r := Rect(DRect.Left, DRect.Top,
              DRect.Left + Round(Picture.Width * ky), DRect.Bottom);
          w := DRect.Right - DRect.Left;
          h := DRect.Bottom - DRect.Top;
          w1 := r.Right - r.Left;
          h1 := r.Bottom - r.Top;
          if (Flags and flPictCenter) <> 0 then
            OffsetRect(r, (w - w1) div 2, (h - h1) div 2);
        end;
        StretchDraw(r, Picture.Graphic);
      end
      else
      begin
        r := DRect;
        if (Flags and flPictCenter) <> 0 then
        begin
          w := DRect.Right - DRect.Left;
          h := DRect.Bottom - DRect.Top;
          OffsetRect(r, (w - Picture.Width) div 2, (h - Picture.Height) div 2);
        end;
        Draw(r.Left, r.Top, Picture.Graphic)
      end;
    end;
    ShowFrame;
  end;
  if DocMode = dmPrinting then CurReport.InternalOnExportData(Self);
end;

procedure TfrPictureView.LoadFromStream(Stream:TStream);
var b: Byte;
    n: Integer;
begin
  inherited LoadFromStream(Stream);
  Stream.Read(b, 1);
  Stream.Read(n, 4);
  case b of
    pkBitmap:   Picture.Assign(TBitmap.Create);
    pkMetafile: Picture.Assign(TMetafile.Create);
    pkIcon:     Picture.Assign(TIcon.Create);
  end;
  if b <> pkNone then
    Picture.Graphic.LoadFromStream(Stream);
  Stream.Seek(n, soFromBeginning);
end;

procedure TfrPictureView.SaveToStream(Stream:TStream);
var b: Byte;
    n,o: Integer;
begin
  inherited SaveToStream(Stream);
  b := pkNone;
  if (Memo.Count = 0) and (Picture.Graphic <> nil) then
    if Picture.Graphic is TBitmap then
      b := pkBitmap
    else if Picture.Graphic is TMetafile then
      b := pkMetafile
    else if Picture.Graphic is TIcon then
      b := pkIcon;
  Stream.Write(b, 1);
  n := Stream.Position;
  Stream.Write(n, 4);
  if b <> pkNone then
    Picture.Graphic.SaveToStream(Stream);
  o := Stream.Position;
  Stream.Seek(n, soFromBeginning);
  Stream.Write(o, 4);
  Stream.Seek(0, soFromEnd);
end;

procedure TfrPictureView.GetBlob(b:TField);
begin
  Picture.Assign(b);
end;

procedure TfrPictureView.DefinePopupMenu(Popup:TPopupMenu);
var m: TMenuItem;
begin
  inherited DefinePopupMenu(Popup);
  m := TMenuItem.Create(Popup);
  m.Caption := frLoadStr('SPictureCenter');
  m.OnClick := P1Click;
  m.Checked := (Flags and flPictCenter) <> 0;
  Popup.Items.Add(m);

  m := TMenuItem.Create(Popup);
  m.Caption := frLoadStr('SKeepAspectRatio');
  m.OnClick := P2Click;
  m.Checked := (Flags and flPictRatio) <> 0;
  Popup.Items.Add(m);
end;

procedure TfrPictureView.P1Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flPictCenter) + Word(Checked) * flPictCenter;
    end;
  end;
  frDesigner.PopupNotify(Self);
end;

procedure TfrPictureView.P2Click(Sender:TObject);
var i: Integer;
    t: TfrView;
begin
  with Sender as TMenuItem do
  begin
    Checked := not Checked;
    for i := 0 to frDesigner.Page.Objects.Count-1 do
    begin
      t := frDesigner.Page.Objects[i];
      if t.Selected then
        t.Flags := (t.Flags and not flPictRatio) + Word(Checked) * flPictRatio;
    end;
  end;
  frDesigner.PopupNotify(Self);
end;

{----------------------------------------------------------------------------}
constructor TfrBand.Create(ATyp:TfrBandType; AParent:TfrPage);
begin
  inherited Create;
  Typ := ATyp;
  Parent := AParent;
  Objects := TList.Create;
  Memo := TStringList.Create;
  GroupValues := TStringList.Create;
  Next := nil;
  Positions[psLocal] := 1;
  Positions[psGlobal] := 1;
  Visible := True;
end;

destructor TfrBand.Destroy;
begin
  if Next <> nil then Next.Free;
  Objects.Free;
  Memo.Free;
  GroupValues.Free;
  if DataSet <> nil then DataSet.Exit;
  inherited Destroy;
end;

procedure TfrBand.InitDataSet(Desc:String);
begin
  if Typ = btGroupHeader then
    GroupCondition := Desc
  else
  begin
    DataSet := CurReport.Owner.FindComponent(Desc) as TfrDataSet;
    if DataSet <> nil then DataSet.Init;
  end;
end;

procedure TfrBand.DoError;
var i: Integer;
begin
  ErrorFlag := True;
  ErrorStr := frLoadStr('SErrorOccured');
  for i := 0 to CurView.Memo.Count-1 do
    ErrorStr := ErrorStr + #13 + CurView.Memo[i];
  ErrorStr := ErrorStr + #13 + frLoadStr('SDoc') + ' ' + CurReport.Name +
    #13 + frLoadStr('SBand') + ' ' + frBandNames[Integer(CurView.Parent.Typ)];
  MasterReport.Terminated := True;
end;

function TfrBand.CalcHeight: Integer;
var i: Integer;
    t: TfrView;
begin
  CurBand := Self;
  SHeight := dy;
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    t.olddy := t.dy;
    if t is TfrMemoView then
      if t.Parent = Self then
      begin
        TfrMemoView(t).DrawMode := drCalcHeight;
        t.Draw(MCanvas);
      end;
  end;
  Result := SHeight;
end;

procedure TfrBand.StretchObjects(MaxHeight:Integer);
var i: Integer;
    t: TfrView;
begin
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    if (t is TfrMemoView) and ((t.Flags and flStretched) <> 0) then
      t.dy := MaxHeight - t.y;
  end;
end;

procedure TfrBand.UnStretchObjects;
var i: Integer;
    t: TfrView;
begin
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    t.dy := t.olddy;
  end;
end;

procedure TfrBand.DrawObject(t:TfrView; DrawMode:TfrDrawMode);
var ox,oy: Integer;
begin
  CurPage := Parent;
  CurBand := Self;
  try
    if (t.Parent = Self) and not DisableDrawing then
    begin
      ox := t.x; Inc(t.x, Parent.XAdjust - Parent.LeftMargin);
      oy := t.y; Inc(t.y, y);
      if t is TfrMemoView then
      begin
        TfrMemoView(t).DrawMode := DrawMode;
        t.Draw(MCanvas);
      end
      else
        t.Draw(MCanvas);
      t.x := ox; t.y := oy;
    end;
  except
    on exception do DoError;
  end;
end;

procedure TfrBand.PrepareSubReports;
var i: Integer;
    t: TfrView;
    Page: TfrPage;
begin
  for i := SubIndex to Objects.Count-1 do
  begin
    t := Objects[i];
    Page := CurReport.Pages[(t as TfrSubReportView).SubPage];
    Page.Mode := pmBuildList;
    Page.FormPage;
    Page.CurY := y + t.y;
    Page.CurBottomY := Parent.CurBottomY;
    Page.XAdjust := Parent.XAdjust + t.x;
    Page.ColCount := 1;
    Page.PlayFrom := 0;
    EOFArr[i - SubIndex] := False;
  end;
  Parent.LastBand := nil;
end;

procedure TfrBand.DoSubReports;
var i: Integer;
    t: TfrView;
    Page: TfrPage;
begin
  repeat
    if not EOFReached then
      for i := SubIndex to Objects.Count-1 do
      begin
        t := Objects[i];
        Page := CurReport.Pages[(t as TfrSubReportView).SubPage];
        Page.CurY := Parent.CurY;
        Page.CurBottomY := Parent.CurBottomY;
      end;
    EOFReached := True;
    MaxY := Parent.CurY;
    for i := SubIndex to Objects.Count-1 do
      if not EOFArr[i - SubIndex] then
      begin
        t := Objects[i];
        Page := CurReport.Pages[(t as TfrSubReportView).SubPage];
        if Page.PlayRecList then
          EOFReached := False
        else
        begin
          EOFArr[i - SubIndex] := True;
          if Page.CurY > MaxY then MaxY := Page.CurY;
        end;
      end;
    if not EOFReached then
    begin
      if Parent.Skip then
      begin
        Parent.LastBand := Self;
        Exit;
      end
      else
        Parent.NewPage;
    end;
  until EOFReached or MasterReport.Terminated;
  for i := SubIndex to Objects.Count-1 do
  begin
    t := Objects[i];
    Page := CurReport.Pages[(t as TfrSubReportView).SubPage];
    Page.ClearRecList;
  end;
  Parent.CurY := MaxY;
  Parent.LastBand := nil;
end;

function TfrBand.DrawObjects(DrawMode:TfrDrawMode): Boolean;
var i: Integer;
    t: TfrView;
begin
  Result := False;
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    if t.Typ = gtSubReport then
    begin
      SubIndex := i;
      Result := True;
      PrepareSubReports;
      DoSubReports;
      break;
    end;
    DrawObject(t, DrawMode);
    if MasterReport.Terminated then break;
  end;
end;

procedure TfrBand.DrawColumnCell(Parnt:TfrBand; CurX:Integer);
var i,sfx,sfy: Integer;
    t: TfrView;
begin
  if DisableDrawing then Exit;
  try
    for i := 0 to Objects.Count-1 do
    begin
      t := Objects[i];
      if Parnt.Objects.IndexOf(t) <> -1 then
      begin
        sfx := t.x; Inc(t.x, CurX);
        sfy := t.y; Inc(t.y, Parnt.y);
        t.Parent := Parnt;
        if t is TfrMemoView then
          TfrMemoView(t).DrawMode := drAll;
        t.Draw(MCanvas);
        t.Parent := Self;
        t.x := sfx;
        t.y := sfy;
      end;
    end;
  except
    on exception do DoError;
  end;
end;

procedure TfrBand.DrawColumnAll;
var Bnd: TfrBand;
    sfpage: Integer;
    CurX: Integer;
  procedure CheckColumnPageBreak;
  var sfDataRow,sfLastDataRow,sfy: Integer;
      b: TfrBand;
  begin
    if CurX + Bnd.dx > Parent.RightMargin then
    begin
      Inc(ColumnXAdjust, CurX - Parent.LeftMargin);
      CurX := Parent.LeftMargin;
      Inc(PageNo);
      if PageNo >= MasterReport.EMFPages.Count then
      begin
        MasterReport.EMFPages.Add(Parent);
        sfDataRow := 0;
        sfLastDataRow := 0;
        if CurReport.FCurrentFilter <> nil then
        begin
          sfDataRow := CurDataRow;
          sfLastDataRow := CurReport.FCurrentFilter.LastDataRow;
        end;
        sfy := Parent.CurY;
        Parent.ShowBand(Parent.Bands[btOverlay]);
        Parent.ShowBand(Parent.Bands[btPageHeader]); // need 'CurY' saving!
        Parent.CurY := sfy;
        if CurReport.FCurrentFilter <> nil then
        begin
          CurDataRow := sfDataRow;
          CurReport.FCurrentFilter.LastDataRow := sfLastDataRow;
        end;
        CurReport.InternalOnProgress(PageNo);
      end
      else
        MCanvas := MasterReport.EMFPages[PageNo].Canvas;
      if Parent.CHCopy then
        if Parent.BandExists(Parent.Bands[btColumnHeader]) then
        begin
          b := Parent.Bands[btColumnHeader];
          b.DrawColumnCell(Self, Parent.LeftMargin);
          CurX := Parent.LeftMargin + b.dx;
        end;
    end;
  end;
begin
  ColumnXAdjust := 0;
  Parent.ColPos := 1;
  CurX := 0;
  sfpage := PageNo;
  if Parent.BandExists(Parent.Bands[btColumnHeader]) then
  begin
    Bnd := Parent.Bands[btColumnHeader];
    Bnd.DrawColumnCell(Self, Bnd.x);
    CurX := Bnd.x + Bnd.dx;
  end;
  if Parent.BandExists(Parent.Bands[btColumnData]) then
  begin
    Bnd := Parent.Bands[btColumnData];
    if CurX = 0 then CurX := Bnd.x;
    Bnd.DataSet.First;
    while not Bnd.DataSet.EOF do
    begin
      CheckColumnPageBreak;
      Bnd.DrawColumnCell(Self, CurX);
      Inc(CurX, Bnd.dx);
      Inc(Parent.ColPos);
      Bnd.DataSet.Next;
      if MasterReport.Terminated then break;
    end;
  end;
  if Parent.BandExists(Parent.Bands[btColumnFooter]) then
  begin
    Bnd := Parent.Bands[btColumnFooter];
    if CurX = 0 then CurX := Bnd.x;
    CheckColumnPageBreak;
    Bnd.DrawColumnCell(Self, CurX);
  end;
  PageNo := sfpage;
  MCanvas := MasterReport.EMFPages[PageNo].Canvas;
  ColumnXAdjust := 0;
end;

function TfrBand.CheckPageBreak(y,dy:Integer; PBreak:Boolean): Boolean;
begin
  Result := False;
  with Parent do
  if y + Bands[btStaticColumnFooter].dy + dy > CurBottomY then
  begin
    if not PBreak then
    begin
      if CurColumn < ColCount-1 then
      begin
        ShowBand(Bands[btStaticColumnFooter]);
        Inc(CurColumn);
        Inc(XAdjust, ColWidth + ColGap);
        CurY := LastStaticColumnY;
        ShowBand(Bands[btStaticColumnHeader]);
      end
      else
        NewPage;
      Inc(CurDataRow);
    end;
    Result := True;
  end;
end;

procedure TfrBand.DrawPageBreak(DrawMode:TfrDrawMode);
var i,th,addy: Integer;
    dy,oldy,olddy,maxy: Integer;
    t: TfrView;
    Flag: Boolean;
  function RoundY: Integer;
  var dy: Integer;
  begin
    dy := t.dy div th * th + t.gapy*2 + 1;
    if dy > t.dy then Dec(dy, th);
    Result := t.dy - dy + 1;
    t.dy := dy;
  end;
  procedure CorrY(t:TfrView; dy:Integer);
  var i: Integer;
      t1: TfrView;
  begin
    for i := 0 to Objects.Count-1 do
    begin
      t1 := Objects[i];
      if t1 <> t then
        if (t1.y > t.y+t.dy) and (t1.x >= t.x) and (t1.x <= t.x+t.dx) then
          Inc(t1.y, dy);
    end;
  end;
begin
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    t.Selected := True;
    t.OriginalRect := Rect(t.x, t.y, t.dx, t.dy);
  end;
  if not CheckPageBreak(y,maxdy,True) then
    DrawObjects(DrawMode)
  else
  begin
    repeat
      dy := Parent.CurBottomY - Parent.Bands[btStaticColumnFooter].dy - y - 2;
      maxy := 0;
      for i := 0 to Objects.Count-1 do
      begin
        t := Objects[i];
        if t.Selected then
        if (t.y >= 0) and (t.y < dy) then
          if (t.y + t.dy < dy) then
          begin
            if maxy < t.y + t.dy then maxy := t.y + t.dy;
            DrawObject(t, DrawMode);
            t.Selected := False;
          end
          else
          begin
            if t is TfrMemoView then
            begin
              th := (t as TfrMemoView).TextHeight;
              olddy := t.dy;
              t.dy := dy - t.y + 1;
              addy := RoundY;
              if t.dy < th then
              begin
                CorrY(t, dy - t.y);
                t.y := dy;
                addy := 0;
              end
              else
              begin
                (t as TfrMemoView).FromLine := 0;
                DrawObject(t, drPart);
                CorrY(t, addy);
              end;
              t.dy := olddy + addy;
            end
            else
              t.y := dy
          end
        else if (t.y < 0) and (t.y + t.dy >= 0) then
          if t.y + t.dy < dy then
          begin
            th := (t as TfrMemoView).TextHeight;
            oldy := t.y; olddy := t.dy;
            t.dy := t.y + t.dy;
            t.y := 0;
            if t.dy > th div 2 then
            begin
              t.dy := (t.Memo1.Count - (t as TfrMemoView).FromLine) * th +
                t.gapy * 2 + 1;
              if maxy < t.y + t.dy then maxy := t.y + t.dy;
              DrawObject(t, drPart);
            end;
            t.y := oldy; t.dy := olddy;
            t.Selected := False;
          end
          else
          begin
            th := (t as TfrMemoView).TextHeight;
            oldy := t.y; olddy := t.dy;
            t.dy := dy;
            t.y := 0;
            addy := RoundY;
            DrawObject(t, drPart);
            t.y := oldy; t.dy := olddy + addy;
          end;
      end;
      Flag := False;
      for i := 0 to Objects.Count-1 do
      begin
        t := Objects[i];
        if t.Selected then Flag := True;
        Dec(t.y, dy);
      end;
      if Flag then CheckPageBreak(y, 10000, False);
      y := Parent.CurY;
    until not Flag;
    maxdy := maxy;
  end;
  for i := 0 to Objects.Count-1 do
  begin
    t := Objects[i];
    t.y := t.OriginalRect.Top;
    t.dy := t.OriginalRect.Bottom;
  end;
  Inc(Parent.CurY, maxdy);
end;

procedure TfrBand.DoDraw;
var sfy,sh: Integer;
    UseY,WasSub: Boolean;
  procedure DrawColumns;
  var i: Integer;
      t: TfrView;
  begin
    for i := 0 to Objects.Count-1 do
    begin
      t := Objects[i];
      if t.Parent <> Self then
      begin
        DrawColumnAll;
        break;
      end;
    end;
  end;

begin
  if Objects.Count = 0 then Exit;
  sfy := y;
  UseY := not (Typ in [btPageFooter, btOverlay, btNone]);
  if UseY then y := Parent.CurY;
  if Stretched then
  begin
    sh := CalcHeight;
    if sh > dy then StretchObjects(sh);
    maxdy := sh;
    if not PageBreak then CheckPageBreak(y, sh, False);
    y := Parent.CurY;
    WasSub := False;
    if PageBreak then
    begin
      DrawPageBreak(drAfterCalcHeight);
      sh := 0;
    end
    else
    begin
      WasSub := DrawObjects(drAfterCalcHeight);
      DrawColumns;
    end;
    UnStretchObjects;
    if not WasSub then Inc(Parent.CurY, sh);
  end
  else
  begin
    if UseY then
    begin
      if not PageBreak then CheckPageBreak(y, dy, False);
      y := Parent.CurY;
    end;
    if PageBreak then
    begin
      maxdy := CalcHeight;
      DrawPageBreak(drAfterCalcHeight);
    end
    else
    begin
      WasSub := DrawObjects(drAll);
      DrawColumns;
      if UseY and not WasSub then Inc(Parent.CurY, dy);
    end;
  end;
  y := sfy;
end;

function TfrBand.DoCalcHeight: Integer;
var b: TfrBand;
begin
  if (Typ in [btMasterData, btDetailData, btSubDetailData]) and
    (Next <> nil) and (Next.Dataset = nil) then
  begin
    b := Self;
    Result := 0;
    repeat
      Result := Result + b.CalcHeight;
      b := b.Next;
    until b = nil;
  end
  else
  begin
    Result := dy;
    if Stretched then Result := CalcHeight;
  end;
end;

function TfrBand.Draw: Boolean;
var b: TfrBand;
begin
  Result := False;
  if not (Typ in [btPageHeader, btPageFooter, btStaticColumnHeader,
    btStaticColumnFooter, btOverlay, btNone]) then
    Inc(CurDataRow)
  else
  begin
    CurDataRow := -1;
    CurReport.InternalOnExportText(0, 0, '', nil);
  end;
  if Typ = btStaticColumnHeader then
    Parent.LastStaticColumnY := Parent.CurY;
  if Typ = btPageFooter then
    y := Parent.CurBottomY;
  if Objects.Count > 0 then
  begin
    if not (Typ in [btPageFooter, btOverlay, btNone]) then
      if (Parent.CurY + DoCalcHeight > Parent.CurBottomY) and not PageBreak then
      begin
        Result := True;
        if Parent.Skip then
          Exit else
          CheckPageBreak(0, 10000, False);
      end;
    EOFReached := True;
    if (Typ in [btMasterData, btDetailData, btSubDetailData]) and
      (Next <> nil) and (Next.Dataset = nil) then
    begin
      b := Self;
      repeat
        b.DoDraw;
        Inc(CurDataRow);
        b := b.Next;
      until b = nil;
    end
    else
    begin
      DoDraw;
      if not (Typ in [btMasterData, btDetailData, btSubDetailData, btGroupHeader]) and
        NewPageAfter then
        Parent.NewPage;
    end;
    if not EOFReached then Result := True;
  end;
end;

procedure TfrBand.InitGroupValues;
var b: TfrBand;
begin
  b := Self;
  while b <> nil do
  begin
    if b.FooterBand <> nil then
    begin
      b.FooterBand.GroupValues.Clear;
      b.FooterBand.Count := 0;
    end;
    b.LastGroupValue := Parser.Calc(b.GroupCondition);
    b := b.Next;
  end;
end;

procedure TfrBand.DoAggregate;
var
  i: Integer;
  t: TfrView;
  FUseHighlight: Boolean;
  s: String;
begin
  for i := 0 to GroupValues.Count - 1 do
  begin
    s := GroupValues[i];
    GroupValues[i] := Copy(s, 1, Pos('=', s) - 1) + '=0' +
      Copy(s, Pos('=', s) + 2, 255);
  end;

  Visible := False;
  CurBand := Self;
  for i := 0 to Objects.Count - 1 do
  begin
    t := Objects[i];
    CurView := t;
    if t is TfrMemoView then
    with t as TfrMemoView do
    begin
      FUseHighlight := UseHighlight;
      ExpandVariables;
      UseHighlight := FUseHighlight;
    end;
  end;
  Visible := True;
  Inc(Count);
end;

{----------------------------------------------------------------------------}
constructor TfrPage.Create(ASize,AWidth,AHeight:Integer; AOr:TPrinterOrientation;
  ADoc:TfrReport);
begin
  inherited Create;
  List := TList.Create;
  Objects := TList.Create;
  ChangePaper(ASize, AWidth, AHeight, AOr);
  PHonFirst := True;
  PrintToPrevPage := False;
  UseMargins := True;
end;

destructor TfrPage.Destroy;
begin
  Clear;
  Objects.Free;
  List.Free;
  inherited Destroy;
end;

procedure TfrPage.ChangePaper(ASize,AWidth,AHeight:Integer; AOr:TPrinterOrientation);
begin
  try
    Prn.SetPrinterInfo(ASize, AWidth, AHeight, AOr);
    Prn.FillPrnInfo(PrnInfo);
  except
    on exception do
    begin
      Prn.SetPrinterInfo($100, AWidth, AHeight, AOr);
      Prn.FillPrnInfo(PrnInfo);
    end;
  end;
  pgSize := Prn.PaperSize;
  pgWidth := Prn.PaperWidth;
  pgHeight := Prn.PaperLength;
  pgOr := Prn.Orientation;
end;

procedure TfrPage.Clear;
var i: Integer;
begin
  for i := Objects.Count-1 downto 0 do
    Delete(i);
end;

procedure TfrPage.Delete(Index:Integer);
begin
  TfrView(Objects[Index]).Free;
  Objects.Delete(Index);
end;

procedure TfrPage.InitReport;
var b: TfrBandType;
begin
  for b := btReportTitle to btNone do
    Bands[b] := TfrBand.Create(b, Self);
  TossObjects;
  InitFlag := True;
  CurPos := 1; ColPos := 1;
end;

procedure TfrPage.DoneReport;
var i: Integer;
    b: TfrBandType;
    t: TfrView;
begin
  if InitFlag then
  begin
    for b := btReportTitle to btNone do
      Bands[b].Free;
    for i := 0 to Objects.Count-1 do // set 'y' value back
    begin
      t := Objects[i];
      if t.Typ <> gtBand then
      begin
        t.y := t.oldy; t.x := t.oldx;
        t.Selected := False;
      end;
    end;
  end;
  InitFlag := False;
end;

function TfrPage.TopMargin: Integer;
begin
  if UseMargins then
    if pgMargins.Top = 0 then
      Result := PrnInfo.Ofy else
      Result := pgMargins.Top
  else
    Result := 0;
end;

function TfrPage.BottomMargin: Integer;
begin
  with PrnInfo do
    if UseMargins then
      if pgMargins.Bottom = 0 then
        Result := Ofy + Ph else
        Result := Pgh - pgMargins.Bottom
    else
      Result := Pgh;
  if BandExists(Bands[btPageFooter]) then
    Result := Result - Bands[btPageFooter].dy;
end;

function TfrPage.LeftMargin: Integer;
begin
  if UseMargins then
    if pgMargins.Left = 0 then
      Result := PrnInfo.Ofx else
      Result := pgMargins.Left
  else
    Result := 0;
end;

function TfrPage.RightMargin: Integer;
begin
  with PrnInfo do
    if UseMargins then
      if pgMargins.Right = 0 then
        Result := Ofx + Pw else
        Result := Pgw - pgMargins.Right
    else
      Result := Pgw;
end;

procedure TfrPage.TossObjects;
var i,j,n,last,miny: Integer;
    b: TfrBandType;
    bt,t: TfrView;
    Bnd,Bnd1: TfrBand;
    FirstBand,Flag: Boolean;
    BArr: Array[0..31] of TfrBand;
begin
  for i := 0 to Objects.Count-1 do // select all objects exclude bands
  begin
    t := Objects[i];
    t.Selected := t.Typ <> gtBand;
    t.oldx := t.x; t.oldy := t.y;
    t.Parent := nil;
    if t.Typ = gtSubReport then
      CurReport.Pages[(t as TfrSubReportView).SubPage].Skip := True;
  end;
  Flag := False;
  for i := 0 to Objects.Count-1 do // search for btColumnXXX bands
  begin
    bt := Objects[i];
    if (bt.Typ = gtBand) and
       (TfrBandType(bt.FrameTyp) in [btColumnHeader..btColumnFooter]) then
    with Bands[TfrBandType(bt.FrameTyp)] do
    begin
      Memo.Assign(bt.Memo);
      x := bt.x; dx := bt.dx;
      InitDataSet(bt.FormatStr);
      Flag := True;
    end;
  end;

  if Flag then // fill a ColumnXXX bands at first
    for b := btColumnHeader to btColumnFooter do
    begin
      Bnd := Bands[b];
      for i := 0 to Objects.Count-1 do
      begin
        t := Objects[i];
        if t.Selected then
         if (t.x >= Bnd.x) and (t.x + t.dx <= Bnd.x + Bnd.dx) then
         begin
           t.x := t.x - Bnd.x;
           t.Parent := Bnd;
           Bnd.Objects.Add(t);
         end;
      end;
    end;

  for b := btReportTitle to btGroupFooter do // fill other bands
  begin
    FirstBand := True;
    Bnd := Bands[b];
    BArr[0] := Bnd;
    Last := 1;
    for i := 0 to Objects.Count-1 do // search for specified band
    begin
      bt := Objects[i];
      if (bt.Typ = gtBand) and (bt.FrameTyp = Integer(b)) then
      begin
        if not FirstBand then
        begin
          Bnd.Next := TfrBand.Create(b,Self);
          Bnd := Bnd.Next;
          BArr[Last] := Bnd;
          Inc(Last);
        end;
        FirstBand := False;
        Bnd.Memo.Assign(bt.Memo);
        Bnd.y := bt.y;
        Bnd.dy := bt.dy;
        with bt as TfrBandView, Bnd do
        begin
          InitDataSet(FormatStr);
          Stretched := (Flags and flStretched) <> 0;
          PrintIfSubsetEmpty := (Flags and flBandPrintIfSubsetEmpty) <> 0;
          if Skip then
          begin
            NewPageAfter := False;
            PageBreak := False;
          end
          else
          begin
            NewPageAfter := (Flags and flBandNewPageAfter) <> 0;
            PageBreak := (Flags and flBandPageBreak) <> 0;
          end;
        end;
        for j := 0 to Objects.Count-1 do // placing objects over band
        begin
          t := Objects[j];
          if (t.Parent = nil) and (t.Typ <> gtSubReport) then
           if t.Selected then
            if (t.y >= Bnd.y) and (t.y < Bnd.y + Bnd.dy - 1) then
            begin
              t.Parent := Bnd;
              t.y := t.y - Bnd.y;
              t.Selected := False;
              Bnd.Objects.Add(t);
            end;
        end;
        for j := 0 to Objects.Count-1 do // placing ColumnXXX objects over band
        begin
          t := Objects[j];
          if t.Parent <> nil then
           if t.Selected then
            if (t.y >= Bnd.y) and (t.y < Bnd.y + Bnd.dy - 1) then
            begin
              t.y := t.y - Bnd.y;
              t.Selected := False;
              Bnd.Objects.Add(t);
            end;
        end;
        for j := 0 to Objects.Count-1 do // placing subreports over band
        begin
          t := Objects[j];
          if (t.Parent = nil) and (t.Typ = gtSubReport) then
           if t.Selected then
            if (t.y >= Bnd.y) and (t.y < Bnd.y + Bnd.dy - 1) then
            begin
              t.Parent := Bnd;
              t.y := t.y - Bnd.y;
              t.Selected := False;
              Bnd.Objects.Add(t);
            end;
        end;
      end;
    end;
    for i := 0 to Last-1 do // sorting bands
    begin
      miny := BArr[i].y; n := i;
      for j := i+1 to Last-1 do
        if BArr[j].y < miny then
        begin
          miny := BArr[j].y;
          n := j;
        end;
      Bnd := BArr[i]; BArr[i] := BArr[n]; BArr[n] := Bnd;
    end;
    Bnd := BArr[0]; Bands[b] := Bnd;
    Bnd.Prev := nil;
    for i := 1 to Last-1 do  // finally ordering
    begin
      Bnd.Next := BArr[i];
      Bnd := Bnd.Next;
      Bnd.Prev := BArr[i-1];
    end;
    Bnd.Next := nil;
    Bands[b].LastBand := Bnd;
  end;

  for i := 0 to Objects.Count-1 do // place other objects on btNone band
  begin
    t := Objects[i];
    if t.Selected then
    begin
      t.Parent := Bands[btNone];
      Bands[btNone].y := 0;
      Bands[btNone].Objects.Add(t);
    end;
  end;

  Bnd := Bands[btGroupHeader].LastBand;
  Bnd1 := Bands[btGroupFooter];
  repeat
    Bnd.FooterBand := Bnd1;
    Bnd := Bnd.Prev;
    Bnd1 := Bnd1.Next;
  until (Bnd = nil) or (Bnd1 = nil);

  if ColCount = 0 then ColCount := 1;
  ColWidth := (RightMargin - LeftMargin - ColCount * ColGap) div ColCount;
end;

procedure TfrPage.ShowBand(b:TfrBand);
begin
  if b <> nil then
    if Mode = pmBuildList then
      AddRecord(b, rtShowBand) else
      b.Draw;
end;

procedure TfrPage.AddRecord(b:TfrBand; rt:TfrBandRecType);
var p: PfrBandRec;
begin
  GetMem(p, SizeOf(TfrBandRec));
  p^.Band := b;
  p^.Action := rt;
  List.Add(p);
end;

procedure TfrPage.ClearRecList;
var i: Integer;
begin
  for i := 0 to List.Count-1 do
    FreeMem(PfrBandRec(List[i]), SizeOf(TfrBandRec));
  List.Clear;
end;

function TfrPage.PlayRecList: Boolean;
var p: PfrBandRec;
    b: TfrBand;
begin
  Result := False;
  while PlayFrom < List.Count do
  begin
    p := List[PlayFrom];
    b := p^.Band;
    case p^.Action of
      rtShowBand:
        begin
          if LastBand <> nil then
          begin
            LastBand.DoSubReports;
            if LastBand <> nil then
            begin
              Result := True;
              Exit;
            end;
          end
          else
            if b.Draw then
            begin
              Result := True;
              Exit;
            end;
        end;
      rtFirst:
        begin
          b.DataSet.First;
          b.Positions[psLocal] := 1;
        end;
      rtNext:
        begin
          b.DataSet.Next;
          Inc(CurPos);
          Inc(b.Positions[psGlobal]);
          Inc(b.Positions[psLocal]);
        end;
    end;
    Inc(PlayFrom);
  end;
end;

procedure TfrPage.DrawPageFooters(FreeCanvases:Boolean);
begin
  CurColumn := 0;
  XAdjust := LeftMargin;
  while (PageNo < MasterReport.EMFPages.Count) and
    (MasterReport.EMFPages[PageNo].Pict <> nil) do
  begin
    MCanvas := MasterReport.EMFPages[PageNo].Canvas;
    if not (Append and WasPF) then
      ShowBand(Bands[btPageFooter]);
    if FreeCanvases then
      MasterReport.EMFPages[PageNo].Canvas.Free;
    Inc(PageNo);
  end;
end;

procedure TfrPage.NewPage;
begin
  CurReport.InternalOnProgress(PageNo + 1);
  ShowBand(Bands[btStaticColumnFooter]);
  DrawPageFooters(True);
  if CurReport.FCurrentFilter <> nil then
  begin
    CurReport.FCurrentFilter.OnEndPage;
    CurReport.FCurrentFilter.OnBeginPage;
  end;
  CurBottomY := BottomMargin;
  MasterReport.EMFPages.Add(Self);
  Append := False;
  ShowBand(Bands[btOverlay]);
  CurY := TopMargin;
  ShowBand(Bands[btPageHeader]);
  ShowBand(Bands[btStaticColumnHeader]);
end;

procedure TfrPage.FormPage;
const
  MAXBNDS = 3;
  Bnds: Array[1..MAXBNDS,TfrBandParts] of TfrBandType =
   ((btMasterHeader,btMasterData,btMasterFooter),
    (btDetailHeader,btDetailData,btDetailFooter),
    (btSubDetailHeader,btSubDetailData,btSubDetailFooter));
var BndStack: Array[1..MAXBNDS*3] of TfrBand;
    MaxLevel,BndStackTop: Integer;
    i,sfPage: Integer;
    HasGroups: Boolean;
  procedure AddToStack(b:TfrBand);
  begin
    if b <> nil then
    begin
      Inc(BndStackTop);
      BndStack[BndStackTop] := b;
    end;
  end;

  procedure ShowStack;
  var i: Integer;
  begin
    for i := 1 to BndStackTop do
      if BandExists(BndStack[i]) then
        ShowBand(BndStack[i]);
    BndStackTop := 0;
  end;

  procedure DoLoop(Level:Integer);
  var WasPrinted: Boolean;
      b,b1,b2: TfrBand;
    procedure InitGroups(b:TfrBand);
    begin
      b.InitGroupValues;
      while b <> nil do
      begin
        Inc(b.Positions[psLocal]);
        Inc(b.Positions[psGlobal]);
        ShowBand(b);
        b := b.Next;
      end;
    end;
  begin
    b := Bands[Bnds[Level,bpData]];
    while (b <> nil) and (b.Dataset <> nil) do
    begin
      b.DataSet.First;
      if Mode = pmBuildList then
        AddRecord(b, rtFirst) else
        b.Positions[psLocal] := 1;

      b1 := Bands[btGroupHeader];
      while b1 <> nil do
      begin
        b1.Positions[psLocal] := 0;
        b1.Positions[psGlobal] := 0;
        b1 := b1.Next;
      end;

      if not b.DataSet.EOF then
      begin
        if (Level = 1) and HasGroups then
          InitGroups(Bands[btGroupHeader]);
        AddToStack(Bands[Bnds[Level,bpHeader]]);
        while not b.DataSet.EOF do
        begin
          Application.ProcessMessages;
          if MasterReport.Terminated then break;
          AddToStack(b);
          WasPrinted := True;
          if Level < MaxLevel then
          begin
            DoLoop(Level + 1);
            if BndStackTop > 0 then
              if b.PrintIfSubsetEmpty then
                ShowStack
              else
              begin
                Dec(BndStackTop);
                WasPrinted := False;
              end;
          end
          else ShowStack;

          if (Level = 1) and HasGroups then
          begin
            b1 := Bands[btGroupFooter];
            while b1 <> nil do
            begin
              b1.DoAggregate;
              b1 := b1.Next;
            end;
          end;

          b.DataSet.Next;

          if (Level = 1) and HasGroups then
          begin
            b1 := Bands[btGroupHeader];
            while b1 <> nil do
            begin
              if (Parser.Calc(b1.GroupCondition) <> b1.LastGroupValue) or
                 b.DataSet.EOF then
              begin
                ShowBand(Bands[Bnds[Level,bpFooter]]);
                b2 := Bands[btGroupHeader].LastBand;
                while b2 <> b1 do
                begin
                  ShowBand(b2.FooterBand);
                  b2.Positions[psLocal] := 0;
                  b2 := b2.Prev;
                end;
                ShowBand(b1.FooterBand);
                if not b.DataSet.EOF then
                begin
                  if b1.NewPageAfter then NewPage;
                  InitGroups(b1);
                  ShowBand(Bands[Bnds[Level,bpHeader]]);
                end;
                b.Positions[psLocal] := 0;
                break;
              end;
              b1 := b1.Next;
            end;
          end;

          if Mode = pmBuildList then
            AddRecord(b, rtNext)
          else if WasPrinted then
          begin
            Inc(CurPos);
            Inc(b.Positions[psGlobal]);
            Inc(b.Positions[psLocal]);
            if not b.DataSet.EOF and b.NewPageAfter then NewPage;
          end;
          if MasterReport.Terminated then break;
        end;
        if BndStackTop = 0 then
          ShowBand(Bands[Bnds[Level,bpFooter]])
        else
          Dec(BndStackTop);
      end;
      b := b.Next;
    end;
  end;

begin
  if Mode = pmNormal then
  begin
    if Append then
      if PrevY = PrevBottomY then
      begin
        Append := False;
        WasPF := False;
        while (PageNo < MasterReport.EMFPages.Count) and
          (MasterReport.EMFPages[PageNo].Pict <> nil) do
        begin
          MasterReport.EMFPages[PageNo].Canvas.Free;
          Inc(PageNo);
        end;
        if CurReport.FCurrentFilter <> nil then
        begin
          CurReport.FCurrentFilter.OnEndPage;
          CurReport.FCurrentFilter.OnBeginPage;
        end;
      end;
    if Append and WasPF then
      CurBottomY := PrevBottomY else
      CurBottomY := BottomMargin;
    if not Append then
    begin
      MasterReport.EMFPages.Add(Self);
      CurY := TopMargin;
      ShowBand(Bands[btOverlay]);
      ShowBand(Bands[btNone]);
    end
    else
      CurY := PrevY;
    CurColumn := 0;
    XAdjust := LeftMargin;
    ShowBand(Bands[btReportTitle]);
    if BandExists(Bands[btPageHeader]) and PHonFirst then
      ShowBand(Bands[btPageHeader]);
    ShowBand(Bands[btStaticColumnHeader]);
  end;

  BndStackTop := 0;
  for i := 1 to MAXBNDS do
    if BandExists(Bands[Bnds[i, bpData]]) then
      MaxLevel := i;
  HasGroups := Bands[btGroupHeader].Objects.Count > 0;
  DoLoop(1);
  if Mode = pmNormal then
  begin
    ShowBand(Bands[btStaticColumnFooter]);
    ShowBand(Bands[btReportSummary]);
    PrevY := CurY;
    PrevBottomY := CurBottomY;
    if CurColumn > 0 then
      PrevY := BottomMargin;
    CurColumn := 0;
    XAdjust := LeftMargin;
    sfPage := PageNo;
    WasPF := False;
    if PFonLast then
    begin
      WasPF := BandExists(Bands[btPageFooter]);
      if WasPF then DrawPageFooters(False);
    end;
    PageNo := sfPage + 1;
  end;
end;

function TfrPage.BandExists(b:TfrBand):Boolean;
begin
  Result := b.Objects.Count > 0;
end;

procedure TfrPage.LoadFromStream(Stream:TStream);
var b: Byte;
begin
  with Stream do
  begin
    Read(pgSize, 4);
    Read(pgWidth, 4);
    Read(pgHeight, 4);
    Read(pgMargins, Sizeof(pgMargins));
    Read(b, 1);
    pgOr := TPrinterOrientation(b);
    Read(PHonFirst, 2);
    Read(PFonLast, 2);
    Read(CHCopy, 2);
    Read(PrintToPrevPage, 2);
    Read(UseMargins, 2);
    Read(ColCount, 4);
    Read(ColGap, 4);
  end;
  ChangePaper(pgSize, pgWidth, pgHeight, pgOr);
end;

procedure TfrPage.SaveToStream(Stream:TStream);
var b: Byte;
begin
  with Stream do
  begin
    Write(pgSize, 4);
    Write(pgWidth, 4);
    Write(pgHeight, 4);
    Write(pgMargins, Sizeof(pgMargins));
    b := Byte(pgOr);
    Write(b, 1);
    Write(PHonFirst, 2);
    Write(PFonLast, 2);
    Write(CHCopy, 2);
    Write(PrintToPrevPage, 2);
    Write(UseMargins, 2);
    Write(ColCount, 4);
    Write(ColGap, 4);
  end;
end;

{-----------------------------------------------------------------------}
constructor TfrPages.Create(AParent:TfrReport);
begin
  inherited Create;
  Parent := AParent;
  FPages := TList.Create;
end;

destructor TfrPages.Destroy;
begin
  Clear;
  FPages.Free;
  inherited Destroy;
end;

function TfrPages.GetCount: Integer;
begin
  Result := FPages.Count;
end;

function TfrPages.GetPages(Index:Integer): TfrPage;
begin
  Result := FPages[Index];
end;

procedure TfrPages.Clear;
var i: Integer;
begin
  for i := 0 to FPages.Count-1 do
    Pages[i].Free;
  FPages.Clear;
end;

procedure TfrPages.Add;
begin
  FPages.Add(TfrPage.Create(9, 0, 0, poPortrait, Parent));
end;

procedure TfrPages.Delete(Index:Integer);
begin
  Pages[Index].Free;
  FPages.Delete(Index);
end;

procedure TfrPages.LoadFromStream(Stream:TStream);
var b: Byte;
    t: TfrView;
procedure AddObject(ot:Byte; clname:String);
begin
  Stream.Read(b,1);
  Pages[b].Objects.Add(frCreateObject(ot, clname));
  t := Pages[b].Objects.Items[Pages[b].Objects.Count-1];
end;
begin
  Clear;
  Stream.Read(Parent.PrintToDefault, 2);
  Stream.Read(Parent.DoublePass, 2);
  frReadMemo(Stream, SMemo);
  if SMemo.Count > 0 then
    Parent.SetPrinterTo(SMemo[0]);
  while Stream.Position < Stream.Size do
  begin
    Stream.Read(b, 1);
    if b = $FF then  // page info
    begin
      Add;
      Pages[Count-1].LoadFromStream(Stream);
    end
    else
    if b = $FE then // values
    begin
      Parent.FVal.ReadBinaryData(Stream);
      frReadMemo(Stream, SMemo);
      Parent.Variables.Assign(SMemo);
    end
    else
    begin
      if b > Integer(gtAddIn) then break;
      if b = gtAddIn then
      begin
        frReadMemo(Stream, SMemo);
        AddObject(gtAddIn, SMemo[0]);
      end
      else
        AddObject(b, '');
      t.LoadFromStream(Stream);
    end;
  end;
end;

procedure TfrPages.SaveToStream(Stream:TStream);
var b: Byte;
    i,j: Integer;
    t: TfrView;
begin
  Stream.Write(Parent.PrintToDefault, 2);
  Stream.Write(Parent.DoublePass, 2);
  SMemo.Clear; SMemo.Add(Prn.Printers[Prn.PrinterIndex]);
  frWriteMemo(Stream, SMemo);
  for i := 0 to Count-1 do // adding pages first
  begin
    b := $FF;
    Stream.Write(b, 1); // page info
    Pages[i].SaveToStream(Stream);
  end;
  for i := 0 to Count-1 do
  begin
    for j := 0 to Pages[i].Objects.Count-1 do // then adding objects
    begin
      t := Pages[i].Objects[j];
      b := Byte(t.Typ);
      Stream.Write(b, 1);
      if t.Typ = gtAddIn then
      begin
        SMemo.Clear; SMemo.Add(t.ClassName);
        frWriteMemo(Stream, SMemo);
      end;
      Stream.Write(i, 1);
      t.SaveToStream(Stream);
    end;
  end;
  b := $FE;
  Stream.Write(b, 1);
  Parent.FVal.WriteBinaryData(Stream);
  SMemo.Assign(Parent.Variables);
  frWriteMemo(Stream, SMemo);
end;

{-----------------------------------------------------------------------}
constructor TfrEMFPages.Create(AParent:TfrReport);
begin
  inherited Create;
  Parent := AParent;
  FPages := TList.Create;
end;

destructor TfrEMFPages.Destroy;
begin
  Clear;
  FPages.Free;
  inherited Destroy;
end;

function TfrEMFPages.GetCount: Integer;
begin
  Result := FPages.Count;
end;

function TfrEMFPages.GetPages(Index:Integer): PfrPageInfo;
begin
  Result := FPages[Index];
end;

procedure TfrEMFPages.Clear;
var i: Integer;
begin
  for i := 0 to FPages.Count-1 do
  begin
    if Pages[i]^.Pict <> nil then Pages[i]^.Pict.Free;
    Pages[i]^.Pict := nil;
    FreeMem(Pages[i], SizeOf(TfrPageInfo));
  end;
  FPages.Clear;
end;

procedure TfrEMFPages.Add(Page:TfrPage);
var p: PfrPageInfo;
begin
  GetMem(p, SizeOf(TfrPageInfo));
  FPages.Add(p);
  with p^ do
  begin
    Pict := TMetafile.Create;
    pgSize := Page.pgSize;
    pgWidth := Page.pgWidth;
    pgHeight := Page.pgHeight;
    pgOr := Page.pgOr;
    pgMargins := Page.UseMargins;
    PrnInfo := Page.PrnInfo;
    Pict.Width := PrnInfo.Pgw;
    Pict.Height := PrnInfo.Pgh;
    Canvas := TMetafileCanvas.Create(Pict, 0);
    MCanvas := Canvas;
{$IFNDEF Delphi2}
    if frCharset <> DEFAULT_CHARSET then
      Canvas.Font.Charset := frCharset;
{$ENDIF}
  end;
end;

procedure TfrEMFPages.LoadFromStream(Stream:TStream);
var i,o,c: Integer;
    b: Byte;
    p: PfrPageInfo;
begin
  Clear;
  frReadMemo(Stream, SMemo);
  if SMemo.Count > 0 then
    Parent.SetPrinterTo(SMemo[0]);
  Stream.Read(c, 4);
  i := 0;
  repeat
    Stream.Read(o, 4);
    GetMem(p, SizeOf(TfrPageInfo));
    FPages.Add(p);
    with p^ do
    begin
      Stream.Read(pgSize, 2);
      Stream.Read(pgWidth, 4);
      Stream.Read(pgHeight, 4);
      Stream.Read(b, 1);
      pgOr := TPrinterOrientation(b);
      Stream.Read(b, 1);
      pgMargins := Boolean(b);
      Pict := TMetaFile.Create;
      Pict.LoadFromStream(Stream);
      Prn.SetPrinterInfo(pgSize, pgWidth, pgHeight, pgOr);
      Prn.FillPrnInfo(PrnInfo);
    end;
    Stream.Seek(o, soFromBeginning);
    Inc(i);
  until i >= c;
end;

procedure TfrEMFPages.SaveToStream(Stream:TStream);
var i,o,n: Integer;
    b: Byte;
begin
  SMemo.Clear; SMemo.Add(Prn.Printers[Prn.PrinterIndex]);
  frWriteMemo(Stream, SMemo);
  n := Count;
  Stream.Write(n, 4);
  i := 0;
  repeat
    o := Stream.Position;
    Stream.Write(o, 4); // dummy write
    with Pages[i]^ do
    begin
      Stream.Write(pgSize, 2);
      Stream.Write(pgWidth, 4);
      Stream.Write(pgHeight, 4);
      b := Byte(pgOr);
      Stream.Write(b, 1);
      b := Byte(pgMargins);
      Stream.Write(b, 1);
      Pict.SaveToStream(Stream);
    end;
    n := Stream.Position;
    Stream.Seek(o, soFromBeginning);
    Stream.Write(n, 4);
    Stream.Seek(0, soFromEnd);
    Inc(i);
  until i >= Count;
end;

{-----------------------------------------------------------------------}
constructor TfrValues.Create;
begin
  inherited Create;
  FItems := TStringList.Create;
end;

destructor TfrValues.Destroy;
begin
  Clear;
  FItems.Free;
  inherited Destroy;
end;

procedure TfrValues.DefineProperties(Filer:TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadBinaryData, WriteBinaryData, FItems.Count > 0);
end;

procedure TfrValues.WriteBinaryData(Stream: TStream);
var i,n: Integer;
procedure WriteStr(s:String);
var c: Integer;
begin
  c := Length(s);
  Stream.WriteBuffer(c, 1);
  Stream.WriteBuffer(PChar(s)^, c);
end;
begin
  with Stream do
  begin
    n := FItems.Count;
    WriteBuffer(n, SizeOf(n));
    for i := 0 to n-1 do
    with Objects[i] do
    begin
      WriteBuffer(Typ, SizeOf(Typ));
      WriteBuffer(OtherKind, SizeOf(OtherKind));
      WriteStr(DataSet);
      WriteStr(Field);
      WriteStr(FItems[i]);
    end;
  end;
end;

procedure TfrValues.ReadBinaryData(Stream: TStream);
var i,j,n: Integer;
function ReadStr:String;
var c: Integer;
begin
  c := 0;
  Stream.ReadBuffer(c, 1);
  SetLength(Result, c);
  Stream.ReadBuffer(PChar(Result)^, c);
end;
begin
  Clear;
  FItems.Sorted := False;
  with Stream do
  begin
    ReadBuffer(n, SizeOf(n));
    for i := 0 to n-1 do
    begin
      j := AddValue;
      with Objects[j] do
      begin
        ReadBuffer(Typ, SizeOf(Typ));
        ReadBuffer(OtherKind, SizeOf(OtherKind));
        DataSet := ReadStr;
        Field := ReadStr;
        FItems[j] := ReadStr;
      end;
    end;
  end;
end;

function TfrValues.GetValue(Index:Integer):TfrValue;
begin
  Result := TfrValue(FItems.Objects[Index]);
end;

function TfrValues.AddValue:Integer;
begin
  Result := FItems.AddObject('', TfrValue.Create);
end;

procedure TfrValues.Clear;
var i: Integer;
begin
  for i := 0 to FItems.Count-1 do
    TfrValue(FItems.Objects[i]).Free;
  FItems.Clear;
end;

function TfrValues.FindVariable(const s:String):TfrValue;
var i: Integer;
begin
  Result := nil;
  i := FItems.IndexOf(s);
  if i <> -1 then
    Result := Objects[i];
end;

{----------------------------------------------------------------------------}
constructor TfrReport.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FPages := TfrPages.Create(Self);
  FEMFPages := TfrEMFPages.Create(Self);
  FVars := TStringList.Create;
  FVal := TfrValues.Create;
  FShowProgress := True;
end;

destructor TfrReport.Destroy;
begin
  FVal.Free;
  FVars.Free;
  FEMFPages.Free;
  FPages.Free;
  inherited Destroy;
end;

procedure TfrReport.DefineProperties(Filer: TFiler);
begin
  inherited DefineProperties(Filer);
  Filer.DefineBinaryProperty('Data', ReadBinaryData, WriteBinaryData, FVars.Count > 0);
end;

procedure TfrReport.WriteBinaryData(Stream: TStream);
var i,n: Integer;
procedure WriteStr(s:String);
var c: Integer;
begin
  c := Length(s);
  Stream.WriteBuffer(c, 1);
  Stream.WriteBuffer(PChar(s)^, c);
end;
begin
  n := FVars.Count;
  Stream.WriteBuffer(n, SizeOf(n));
  for i := 0 to n-1 do
    WriteStr(FVars[i]);
end;

procedure TfrReport.ReadBinaryData(Stream: TStream);
var i,n: Integer;
function ReadStr:String;
var c: Integer;
begin
  c := 0;
  Stream.ReadBuffer(c, 1);
  SetLength(Result, c);
  Stream.ReadBuffer(PChar(Result)^, c);
end;
begin
  FVars.Clear;
  Stream.ReadBuffer(n, SizeOf(n));
  for i := 0 to n-1 do
    FVars.Add(ReadStr);
end;

// report building events
procedure TfrReport.InternalOnProgress(Percent: Integer);
begin
  if Assigned(FOnProgress) then
    FOnProgress(Percent)
  else if FShowProgress then
  with ProgressForm do
  begin
    if MasterReport.DoublePass and MasterReport.FinalPass then
      Label1.Caption := FirstCaption + '  ' + IntToStr(Percent) + ' ' +
        frLoadStr('SFrom') + ' ' + IntToStr(SavedAllPages)
    else
      Label1.Caption := FirstCaption + '  ' + IntToStr(Percent);
    Application.ProcessMessages;
  end;
end;

procedure TfrReport.InternalOnGetValue(const ParName:String; var ParValue:String);
begin
  SubValue := '';
  CurVariable := ParName;
  GetVariableValue(ParName, CurValue);
  if CurView.Typ = gtMemo then
    with CurView as TfrMemoView do
      if UseHighlight then
        if Parser.Calc(HighlightStr) = 0 then
          UseHighlight := False;
  ParValue := FormatValue(CurValue, CurView.Format, CurView.FormatStr);
end;

procedure TfrReport.InternalOnEnterRect(Memo:TStringList; View:TfrView);
var Value: TfrValue;
    s: String;
    i: Integer;
    DataSet: TDataSet;
    Field: TField;
begin
  if HasBlobs then
    if Memo.Count > 0 then
    begin
      s := Memo[0];
      i := Length(s);
      if (i > 2) and (s[1] = '[') then
      begin
        while (i > 0) and (s[i] <> ']') do Dec(i);
        s := Copy(s, 2, i-2);
        DataSet := nil; Field := nil;
        Value := Values.FindVariable(s);
        if Value = nil then
        begin
          DataSet := GetDefaultDataset;
          frGetDatasetAndField(s, DataSet, Field);
        end
        else if Value.Typ = vtDBField then
          Field := Value.DField;
        if (Field <> nil) and (Field.DataType in [ftBlob..ftTypedBinary]) then
          View.GetBlob(Field);
      end;
    end;
  if Assigned(FOnEnterRect) then FOnEnterRect(Memo, View);
end;

procedure TfrReport.InternalOnExportData(View:TfrView);
begin
  if FCurrentFilter <> nil then
    FCurrentFilter.OnData(CurView.x + ColumnXAdjust, CurView.y, View);
end;

procedure TfrReport.InternalOnExportText(x,y:Integer; const text:String; View:TfrView);
begin
  if (DocMode = dmPrinting) and (FCurrentFilter <> nil) then
  with FCurrentFilter do
  begin
    DataRow := CurDataRow;
    TextRow := CurStrNo;
    RowTop := CurView.Parent.y;
    FontHeight := CurFontHeight;
    OnText(x + ColumnXAdjust, y, text, View);
  end;
end;

function TfrReport.FormatValue(V:Variant; Format:Integer;
  const FormatStr:String):String;
const
  FormatArr: Array[1..2,0..3] of String[12] =
    (('dd.mm.yy','dd.mm.yyyy','d mmm yyyy','d mmmm yyyy'),
    ('hh:nn:ss','h:nn:ss','hh:nn','h:nn'));
  BoolStr: Array[1..3,0..1] of string[3] = (('',''),('0','1'),(' ','X'));
var f1,f2: Integer;
    c: Char;
begin
  if v = Null then
  begin
    Result := '';
    Exit;
  end;
  c := DecimalSeparator;
  f1 := (Format div $01000000) and $0F;
  f2 := (Format div $00010000) and $FF;
  try
    case f1 of
      0: Result := v;
      1:
        begin
          DecimalSeparator := Chr(Format and $FF);
          case f2 of
            0: Result := FormatFloat('###.##', v);
            1: Result := FloatToStrF(v, ffFixed, 10, (Format div $0100) and $FF);
            2: Result := FormatFloat('#,###.##', v);
            3: Result := FloatToStrF(v, ffNumber, 10, (Format div $0100) and $FF);
            4: Result := FormatFloat(FormatStr, v);
          end;
        end;
      2: if f2 = 4 then
           Result := FormatDateTime(FormatStr, v)
         else
           Result := FormatDateTime(FormatArr[1,f2], v);
      3: if f2 = 4 then
           Result := FormatDateTime(FormatStr, v)
         else
           Result := FormatDateTime(FormatArr[2,f2], v);
      4:
         begin
           if f2 = 1 then
             if Integer(v) = 0 then
               Result := frLoadStr('SNo') else
               Result := frLoadStr('SYes')
           else
             Result := BoolStr[f2,Integer(v)];
         end;
    end;
  except
    on exception do Result := v;
  end;
  DecimalSeparator := c;
end;

procedure TfrReport.GetVariableValue(const s:String; var v:Variant);
var Value: TfrValue;
    DataSet: TDataSet;
    Field: TField;
begin
  v := Null;
  if Assigned(FOnGetValue) then FOnGetValue(s, v);
  if v = Null then
  begin
    Value := Values.FindVariable(s);
    if Value <> nil then
      with Value do
      case Typ of
        vtNotAssigned:
          v := '';
        vtDBField:
          v := DField.AsVariant;
        vtOther:
          if OtherKind = 1 then
            v := Parser.Calc(Field) else
            v := Parser.Calc(frSpecFuncs[OtherKind]);
      end
    else
    begin
      DataSet := GetDefaultDataSet;
      frGetDataSetAndField(s, DataSet, Field);
      if Field <> nil then
        v := Field.AsVariant
      else if s = 'VALUE' then
        v := CurValue
      else if s = frSpecFuncs[0] then
        v := PageNo + 1
      else if s = frSpecFuncs[2] then
        v := CurDate
      else if s = frSpecFuncs[3] then
        v := CurTime
      else if s = frSpecFuncs[4] then
        v := CurBand.Positions[psLocal]
      else if s = frSpecFuncs[5] then
        v := CurBand.Positions[psGlobal]
      else if s = frSpecFuncs[6] then
        v := CurPage.ColPos
      else if s = frSpecFuncs[7] then
        v := CurPage.CurPos
      else if s = frSpecFuncs[8] then
        v := SavedAllPages
      else
      begin
        if s <> SubValue then
        begin
          SubValue := s;
          v := Parser.Calc(s);
          SubValue := '';
        end
        else raise(EParserError.Create(''));
      end;
    end;
  end;
end;

procedure TfrReport.OnGetParsFunction(const name:String; p1,p2,p3:Variant;
   var val:String);
var i: Integer;
begin
  for i := 0 to frFunctionsCount-1 do
    if frFunctions[i].FunctionLibrary.OnFunction(name, p1, p2, p3, val) then
      exit;
  if Assigned(FOnFunction) then FOnFunction(name, p1, p2, p3, val);
end;

// load/save methods
procedure TfrReport.LoadFromStream(Stream:TStream);
begin
  Stream.Read(frVersion, 1);
  if frVersion < 21 then
  begin
    frVersion := 21;
    Stream.Position := 0;
  end;
  Pages.LoadFromStream(Stream);
end;

procedure TfrReport.SaveToStream(Stream:TStream);
begin
  frVersion := frCurrentVersion;
  Stream.Write(frVersion, 1);
  Pages.SaveToStream(Stream);
end;

procedure TfrReport.LoadFromFile(FName:String);
var Stream: TFileStream;
begin
  Stream := TFileStream.Create(FName, fmOpenRead);
  LoadFromStream(Stream);
  Stream.Free;
end;

procedure TfrReport.SaveToFile(FName:String);
var Stream: TFileStream;
begin
  Stream := TFileStream.Create(FName, fmCreate);
  SaveToStream(Stream);
  Stream.Free;
end;

{$IFDEF StoreToDB}
procedure TfrReport.LoadFromDB(Table:TTable; DocN:Integer);
var Stream: TBlobStream;
begin
  if Table.FindKey([DocN]) then
  begin
    Stream := TBlobStream.Create((Table.Fields[1] as TBlobField), bmRead);
    LoadFromStream(Stream);
    Stream.Free;
  end;
end;

procedure TfrReport.SaveToDB(Table:TTable; DocN:Integer);
var Stream: TBlobStream;
begin
  if Table.FindKey([DocN]) then Table.Delete;
  Table.Append;
  Table.Fields[0].AsInteger := DocN;
  Stream := TBlobStream.Create((Table.Fields[1] as TBlobField), bmWrite);
  SaveToStream(Stream);
  Stream.Free;
  Table.Post;
end;
{$ENDIF}

procedure TfrReport.LoadPreparedReport(FName:String);
var Stream: TFileStream;
begin
  Stream := TFileStream.Create(FName, fmOpenRead);
  EMFPages.LoadFromStream(Stream);
  Stream.Free;
  CanRebuild := False;
end;

procedure TfrReport.SavePreparedReport(FName:String);
var Stream: TFileStream;
begin
  Stream := TFileStream.Create(FName, fmCreate);
  EMFPages.SaveToStream(Stream);
  Stream.Free;
end;

procedure TfrReport.LoadTemplate(FName:String; comm:TStrings;
  Bmp:TBitmap; Load:Boolean);
var Stream: TFileStream;
    b: Byte;
    fb: TBitmap;
    fm: TStringList;
    pos: Integer;
begin
  fb := TBitmap.Create;
  fm := TStringList.Create;
  Stream := TFileStream.Create(FName, fmOpenRead);
  if Load then
  begin
    frReadMemo(Stream, fm);
    Stream.Read(pos, 4);
    Stream.Read(b, 1);
    if b <> 0 then
      fb.LoadFromStream(Stream);
    Stream.Position := pos;
    Pages.LoadFromStream(Stream);
  end
  else
  begin
    frReadMemo(Stream, Comm);
    Stream.Read(pos, 4);
    Bmp.Assign(nil);
    Stream.Read(b, 1);
    if b <> 0 then
      Bmp.LoadFromStream(Stream);
  end;
  fm.Free; fb.Free;
  Stream.Free;
end;

procedure TfrReport.SaveTemplate(FName:String; Comm:TStrings; Bmp:TBitmap);
var Stream: TFileStream;
    b: Byte;
    pos,lpos: Integer;
begin
  Stream := TFileStream.Create(FName, fmCreate);
  frWriteMemo(Stream, Comm);
  b := 0;
  pos := Stream.Position;
  lpos := 0;
  Stream.Write(lpos, 4);
  if Bmp.Empty then
    Stream.Write(b, 1)
  else
  begin
    b := 1;
    Stream.Write(b, 1);
    Bmp.SaveToStream(Stream);
  end;
  lpos := Stream.Position;
  Stream.Position := pos;
  Stream.Write(lpos, 4);
  Stream.Position := lpos;
  Pages.SaveToStream(Stream);
  Stream.Free;
end;

// report manipulation methods
procedure TfrReport.DesignReport;
begin
  if Pages.Count = 0 then Pages.Add;
  CurReport := Self;
  if frDesigner <> nil then
    frDesigner.ShowModal;
end;

var SavedFilter: TfrExportFilter;
    FirstPassTerminated: Boolean;

procedure TfrReport.BuildBeforeModal(Sender:TObject);
begin
  DoBuildReport;
  if FinalPass then
    if Terminated then
      ProgressForm.ModalResult := mrCancel else
      ProgressForm.ModalResult := mrOk
  else
  begin
    FirstPassTerminated := Terminated;
    FCurrentFilter := SavedFilter;
    SavedAllPages := EMFPages.Count;
    DoublePass := False;
    PrepareReport; // do final pass
    DoublePass := True;
  end;
end;

function TfrReport.PrepareReport: Boolean;
var s: String;
begin
  s := frLoadStr('SReportPreparing');
  Terminated := False;
  Result := True;
  Append := False;
  CurDate := Date; CurTime := Time;
  MasterReport := Self;
  DisableDrawing := False;
  FinalPass := True;
  PageNo := 0;
  EMFPages.Clear;
  if DoublePass then
  begin
    DisableDrawing := True;
    SavedFilter := FCurrentFilter; FCurrentFilter := nil;
    FinalPass := False;
    if not Assigned(FOnProgress) and FShowProgress then
      with ProgressForm do
      begin
        if Title = '' then
          Caption := s else
          Caption := s + ' - ' + Title;
        FirstCaption := frLoadStr('SFirstPass');
        Label1.Caption := FirstCaption + '  1';
        OnBeforeModal := BuildBeforeModal;
        Show_Modal(Self);
      end
    else
      BuildBeforeModal(nil);
    Exit;
  end;
  if FCurrentFilter <> nil then FCurrentFilter.OnBeginDoc;
  if not Assigned(FOnProgress) and FShowProgress then
    with ProgressForm do
    begin
      if Title = '' then
        Caption := s else
        Caption := s + ' - ' + Title;
      FirstCaption := frLoadStr('SPagePreparing');
      Label1.Caption := FirstCaption + '  1';
      OnBeforeModal := BuildBeforeModal;
      if Visible then
      begin
        if not FirstPassTerminated then DoublePass := True;
        BuildBeforeModal(nil);
      end
      else
      begin
        SavedAllPages := 0;
        if Show_Modal(Self) = mrCancel then
          Result := False;
      end;
    end
  else
    BuildBeforeModal(nil);
  if FCurrentFilter <> nil then FCurrentFilter.OnEndDoc;
  Terminated := False;
end;

procedure TfrReport.ExportTo(Filter:TClass; FileName:String);
var Stream: TFileStream;
    Opt: Boolean;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  FCurrentFilter := TfrExportFilter(Filter.NewInstance);
  FCurrentFilter.Create(Stream);
  Opt := DoublePass;
  DoublePass := False;
  PrepareReport;
  DoublePass := Opt;
  FCurrentFilter.Free;
  FCurrentFilter := nil;
  Stream.Free;
end;

procedure TfrReport.DoBuildReport;
var i: Integer;
begin
  CanRebuild := True;
  DocMode := dmPrinting;
  Values.Items.Sorted := True;
  StartReport;
  InternalOnProgress(PageNo+1);
  for i := 0 to Pages.Count-1 do
    Pages[i].Skip := False;
  for i := 0 to Pages.Count-1 do
    Pages[i].InitReport;
  for i := 0 to Pages.Count-1 do
  begin
    FCurPage := Pages[i];
    if FCurPage.Skip then continue;
    if Assigned(FOnBeginPage) then
      FOnBeginPage(i, FCurPage.Bands[btMasterData].Memo,
                  FCurPage.Bands[btDetailData].Memo,
                  FCurPage.Bands[btSubDetailData].Memo);
    if (FCurrentFilter <> nil) and not Append then FCurrentFilter.OnBeginPage;

    FCurPage.Mode := pmNormal;
    FCurPage.FormPage;
    if Assigned(FOnEndPage) then FOnEndPage(1);

    Append := False;
    if ((i = Pages.Count-1) and CompositeMode) or
       ((i <> Pages.Count-1) and Pages[i+1].PrintToPrevPage) then
    begin
      Dec(PageNo);
      Append := True;
    end;
    if not Append then
    begin
      Dec(PageNo);
      while (PageNo < MasterReport.EMFPages.Count) and
       (MasterReport.EMFPages[PageNo].Pict <> nil) do
      begin
        MasterReport.EMFPages[PageNo].Canvas.Free;
        Inc(PageNo);
      end;
      InternalOnProgress(PageNo);
      if FCurrentFilter <> nil then FCurrentFilter.OnEndPage;
    end;
    if MasterReport.Terminated then break;
  end;
  if Assigned(FOnEndDoc) then FOnEndDoc;
  for i := 0 to Pages.Count-1 do
    Pages[i].DoneReport;
  InternalOnProgress(PageNo);
  Values.Items.Sorted := False;
end;

procedure TfrReport.ShowReport;
begin
  PrepareReport;
  if ErrorFlag then
    MessageBox(0, PChar(ErrorStr), PChar(frLoadStr('SError')),
      mb_Ok+mb_IconError) else
    ShowPreparedReport;
  EMFPages.Clear;
end;

procedure TfrReport.ShowPreparedReport;
var s: String;
begin
  s := frLoadStr('SPreview');
  if Title <> '' then s := s + ' - ' + Title;
  PreviewForm.Caption := s;
  PreviewForm.Show_Modal(Self);
end;

procedure TfrReport.PrintBeforeModal(Sender:TObject);
begin
  DoPrintReport(FFromPage, FToPage, Fn, FCollate);
  ProgressForm.ModalResult := mrOk;
end;

procedure TfrReport.PrintPreparedReport(FromPage,ToPage,n:Integer; Collate:Boolean);
var s: String;
begin
  s := frLoadStr('SReportPreparing');
  Terminated := False;
  FFromPage := FromPage; FToPage := ToPage; Fn := n; FCollate := Collate;
  if not Assigned(FOnProgress) and FShowProgress then
    with ProgressForm do
    begin
      if Title = '' then
        Caption := s else
        Caption := s + ' - ' + Title;
      FirstCaption := frLoadStr('SPagePrinting');
      Label1.Caption := FirstCaption;
      OnBeforeModal := PrintBeforeModal;
      Show_Modal(Self);
    end
  else
    PrintBeforeModal(nil);
  Terminated := False;
end;

procedure TfrReport.DoPrintReport(FromPage,ToPage,n:Integer; Collate:Boolean);
var i,j,nPages: Integer;
    f: Boolean;
procedure PrintPage(n:Integer);
begin
  with Printer, EMFPages[n]^ do
  begin
    if not Prn.IsEqual(pgSize, pgWidth, pgHeight, pgOr) then
    begin
      EndDoc;
      Prn.SetPrinterInfo(pgSize, pgWidth, pgHeight, pgOr);
      BeginDoc;
    end
    else if not f then NewPage;
    Prn.FillPrnInfo(PrnInfo);
    with PrnInfo do
      if pgMargins then
        Printer.Canvas.StretchDraw(
          Rect(-POfx, -POfy, PPgw - POfx, PPgh - POfy), Pict)
      else
        Printer.Canvas.StretchDraw(Rect(0, 0, PPw, PPh), Pict);
  end;
  InternalOnProgress(n + 1);
  Application.ProcessMessages;
  f := False;
end;

begin
  Prn.Printer := Printer;
  Dec(FromPage); Dec(ToPage);
  if ToPage >= EMFPages.Count then ToPage := EMFPages.Count-1;
  nPages := ToPage - FromPage + 1;
  with EMFPages[0]^ do
  begin
    Prn.SetPrinterInfo(pgSize, pgWidth, pgHeight, pgOr);
    Prn.FillPrnInfo(PrnInfo);
  end;
  if Title <> '' then
    Printer.Title := 'FastReport: ' + Title else
    Printer.Title := 'FastReport: ' + frLoadStr('SRepNoname');
  Printer.BeginDoc;
  f := True;
  if Collate then
    for i := 0 to n-1 do
     for j := 0 to nPages-1 do
     begin
       PrintPage(j + FromPage);
       if Assigned(OnProgress) then
         OnProgress((i * nPages + j + 1) * 100 div (nPages * n));
       if Terminated then
       begin
         Printer.Abort;
         exit;
       end;
     end
  else
  begin
    for i := 0 to nPages-1 do
     for j := 0 to n-1 do
     begin
       PrintPage(i + FromPage);
       if Assigned(OnProgress) then
         OnProgress((i * n + j + 1) * 100 div (nPages * n));
       if Terminated then
       begin
         Printer.Abort;
         exit;
       end;
     end
  end;
  Printer.EndDoc;
end;

// printer manipulation methods

procedure TfrReport.SetPrinterTo(PrnName:String);
begin
  if not PrintToDefault then
    if Prn.Printers.IndexOf(PrnName) <> -1 then
      Prn.PrinterIndex := Prn.Printers.IndexOf(PrnName);
end;

function TfrReport.ChangePrinter(OldIndex,NewIndex:Integer):Boolean;
procedure ChangePages;
var i: Integer;
begin
  for i := 0 to Pages.Count-1 do
    with Pages[i] do
      ChangePaper(pgSize, pgWidth, pgHeight, pgOr);
end;
begin
  Result := True;
  try
    Prn.PrinterIndex := NewIndex;
    Prn.PaperSize := -1;
    ChangePages;
  except
    on Exception do
    begin
      MessageBox(0, PChar(frLoadStr('SPrinterError')),
        PChar(frLoadStr('SError')), mb_IconError + mb_Ok);
      Prn.PrinterIndex := OldIndex;
      ChangePages;
      Result := False;
    end;
  end;
end;

// miscellaneous methods
procedure TfrReport.PrepareDataSets;
var i: Integer;
begin
  HasBlobs := False;
  with Values do
  for i := 0 to Items.Count-1 do
    with Objects[i] do
    if Typ = vtDBField then
    begin
      DSet := frGetDataSet(DataSet);
      DField := DSet.FieldByName(Field);
      if DField.DataType in [ftBlob..ftTypedBinary] then HasBlobs := True;
    end;
end;

procedure TfrReport.StartReport;
begin
  CurReport := Self;
  if Assigned(FOnBeginDoc) then FOnBeginDoc;
  Parser.OnGetValue := GetVariableValue;
  Parser.OnFunction := OnGetParsFunction;
  PrepareDataSets;
  ErrorFlag := False;
end;

procedure TfrReport.SetVars(Value:TStrings);
begin
  FVars.Assign(Value);
end;

procedure TfrReport.GetVarList(CatNo:Integer; List:TStrings);
var i,n: Integer;
    s: String;
begin
  List.Clear;
  i := 0; n := 0;
  if FVars.Count > 0 then
    repeat
      s := FVars[i];
      if Length(s) > 0 then
        if s[1] <> ' ' then Inc(n);
      Inc(i);
    until n > CatNo;
  while i < FVars.Count do
  begin
    s := FVars[i];
    if (s <> '') and (s[1] = ' ') then
      List.Add(Copy(s, 2, Length(s)-1))
    else
      break;
    Inc(i);
  end;
end;

procedure TfrReport.GetCategoryList(List:TStrings);
var i: Integer;
    s: String;
begin
  List.Clear;
  for i := 0 to FVars.Count-1 do
  begin
    s := FVars[i];
    if (s <> '') and (s[1] <> ' ') then List.Add(s);
  end;
end;

function TfrReport.FindVariable(Variable:String):Integer;
var i: Integer;
begin
  Result := -1;
  Variable := ' ' + Variable;
  for i := 0 to FVars.Count-1 do
    if Variable = FVars[i] then
    begin
      Result := i;
      break;
    end;
end;

{----------------------------------------------------------------------------}
constructor TfrCompositeReport.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  Reports := TList.Create;
end;

destructor TfrCompositeReport.Destroy;
begin
  Reports.Free;
  inherited Destroy;
end;

procedure TfrCompositeReport.DoBuildReport;
var i: Integer;
    Doc: TfrReport;
begin
  CanRebuild := True;
  PageNo := 0;
  if Assigned(FOnBeginDoc) then FOnBeginDoc;
  for i := 0 to Reports.Count-1 do
  begin
    Doc := TfrReport(Reports[i]);
    CompositeMode := False;
    if i <> Reports.Count-1 then
      if (TfrReport(Reports[i+1]).Pages.Count > 0) and
        TfrReport(Reports[i+1]).Pages[0].PrintToPrevPage then
        CompositeMode := True;
    Doc.FCurrentFilter := FCurrentFilter;
    Doc.DoBuildReport;
    Doc.FCurrentFilter := nil;
    Append := CompositeMode;
    CompositeMode := False;
    if Terminated then break;
  end;
  if Assigned(FOnEndDoc) then FOnEndDoc;
end;


{----------------------------------------------------------------------------}
procedure TfrObjEditorForm.ShowEditor(t:TfrView);
begin
// abstract method
end;


{----------------------------------------------------------------------------}
constructor TfrExportFilter.Create(AStream:TStream);
begin
  inherited Create;
  Stream := AStream;
end;

procedure TfrExportFilter.OnBeginDoc;
begin
// abstract method
end;

procedure TfrExportFilter.OnEndDoc;
begin
// abstract method
end;

procedure TfrExportFilter.OnBeginPage;
begin
// abstract method
end;

procedure TfrExportFilter.OnEndPage;
begin
// abstract method
end;

procedure TfrExportFilter.OnData(x,y:Integer; View:TfrView);
begin
// abstract method
end;

procedure TfrExportFilter.OnText(x,y:Integer; const text:String; View:TfrView);
begin
// abstract method
end;


{----------------------------------------------------------------------------}
constructor TfrFunctionLibrary.Create;
begin
  inherited Create;
  List := TStringList.Create;
  List.Sorted := True;
end;

destructor TfrFunctionLibrary.Destroy;
begin
  List.Free;
  inherited Destroy;
end;

function TfrFunctionLibrary.OnFunction(const FName:String; p1,p2,p3:Variant;
  var val:String): Boolean;
var i: Integer;
begin
  Result := False;
  if List.Find(FName, i) then
  begin
    DoFunction(i, p1, p2, p3, val);
    Result := True;
  end;
end;


{----------------------------------------------------------------------------}
constructor TfrStdFunctionLibrary.Create;
begin
  inherited Create;
  with List do
  begin
    Add('AVG');
    Add('COUNT');
    Add('FORMATDATETIME');
    Add('FORMATFLOAT');
    Add('LOWERCASE');
    Add('MAX');
    Add('MIN');
    Add('NAMECASE');
    Add('STRTODATE');
    Add('STRTOTIME');
    Add('SUM');
    Add('UPPERCASE');
  end;
end;

procedure TfrStdFunctionLibrary.DoFunction(FNo:Integer; p1,p2,p3:Variant;
  var val:String);
var DataSet: TDataSet;
    Field: TField;
    s1: String;
    min,max,avg,sum,count,d,v: Double;
    dk: (dkNone,dkSum,dkMin,dkMax,dkAvg,dkCount);
begin
  dk := dkNone;
  case FNo of
    0: dk := dkAvg;
    1: dk := dkCount;
    2: val := '''' + FormatDateTime(Parser.Calc(p1), Parser.Calc(p2)) + '''';
    3: val := '''' + FormatFloat(Parser.Calc(p1), Parser.Calc(p2)) + '''';
    4: val := '''' + AnsiLowerCase(Parser.Calc(p1)) + '''';
    5: dk := dkMax;
    6: dk := dkMin;
    7:
      begin
        s1 := AnsiLowerCase(Parser.Calc(p1));
        if Length(s1) > 0 then
          val := '''' + AnsiUpperCase(s1[1]) + Copy(s1, 2, Length(s1)-1) + ''''
        else
          val := '''' + '''';
      end;
    8: val := '%d''' + Parser.Calc(p1) + '''';
    9: val := '%t''' + Parser.Calc(p1) + '''';
    10: dk := dkSum;
    11: val := '''' + AnsiUpperCase(Parser.Calc(p1)) + '''';
  end;

  if dk <> dkNone then
  begin
    if (CurBand.Typ = btGroupFooter) and not CurBand.Visible then
    begin
      val := '0';
      s1 := CurBand.GroupValues.Values[List[FNo] + p1];
      if s1 <> '' then
        if s1[1] = '1' then
          Exit else
          s1 := Copy(s1, 2, 255);
      d := 0;
      if dk <> dkCount then d := Parser.Calc(p1);
      if s1 = '' then
        if dk = dkMin then s1 := '1e200'
        else if dk = dkMax then s1 := '-1e200'
        else s1 := '0';
      v := StrToFloat(s1);
      case dk of
        dkAvg: v := v + d;
        dkCount: v := v + 1;
        dkMax: if v < d then v := d;
        dkMin: if v > d then v := d;
        dkSum: v := v + d;
      end;
      CurBand.GroupValues.Values[List[FNo] + p1] := '1' + FloatToStr(v);
      Exit;
    end;
    if (CurBand.Typ = btGroupFooter) and CurBand.Visible then
    begin
      val := Copy(CurBand.GroupValues.Values[List[FNo] + p1], 2, 255);
      if dk = dkAvg then
        val := FloatToStr(StrToFloat(val) / CurBand.Count);
      Exit;
    end;

    DataSet := GetDefaultDataSet;
    if dk = dkCount then
      DataSet := frGetDataSet(p1) else
      frGetDataSetAndField(p1, DataSet, Field);
    min := 1e200; max := -1e200; sum := 0; count := 0; avg := 0;
    DataSet.First;
    while not DataSet.EOF do
    begin
      v := 0;
      if dk <> dkCount then
        v := Field.AsFloat;
      if v > max then max := v;
      if v < min then min := v;
      sum := sum + v;
      count := count + 1;
      DataSet.Next;
    end;
    if count > 0 then
      avg := sum / count;
    d := 0;
    case dk of
      dkSum: d := sum;
      dkMin: d := min;
      dkMax: d := max;
      dkAvg: d := avg;
      dkCount: d := count;
    end;
    val := FloatToStr(d);
  end;
end;


{----------------------------------------------------------------------------}
procedure DoInit;
const Clr: Array[0..1] of TColor = (clWhite,clSilver);
var i,j: Integer;
begin
  SMemo := TStringList.Create;
  SBmp := TBitmap.Create;
  SBmp.Width := 8; SBmp.Height := 8;
  for j := 0 to 7 do
    for i := 0 to 7 do
      SBmp.Canvas.Pixels[i,j] := Clr[(j + i) mod 2];
  frRegisterFunctionLibrary(TfrStdFunctionLibrary);
  PreviewForm := TPreviewForm.Create(nil);
  ProgressForm := TProgressForm.Create(nil);
  frCharset := DEFAULT_CHARSET;

  for i := 0 to 21 do
    frBandNames[i] := frLoadStr('SBand' + IntToStr(i+1));
  for i := 0 to frSpecCount-1 do
    frSpecArr[i] := frLoadStr('SVar' + IntToStr(i+1));
end;

procedure DoExit;
var i: Integer;
begin
  SBmp.Free;
  SMemo.Free;
  ProgressForm.Free;
  PreviewForm.Free;
  for i := 0 to frFunctionsCount-1 do
    frFunctions[i].FunctionLibrary.Free;
end;


initialization
  DoInit;

finalization
  DoExit;

end.
