
{*****************************************}
{                                         }
{             FastReport v2.2             }
{            Report DB dataset            }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_DBSet;

interface

{$I FR_Vers.inc}

uses
  SysUtils, Windows, Messages, Classes, DB, FR_DSet;

type
  TfrDBDataSet = class(TfrDataset)
  private
    FDataSource: TDataSource;
    FOpenDataSource,FCloseDataSource: Boolean;
    FOnOpen,FOnClose: TNotifyEvent;
    FBookmark: TBookmark;
    FEof: Boolean;
    function DataSet: TDataSet;
  protected
    procedure Notification(AComponent: TComponent;
                           Operation: TOperation); override;
  public
    constructor Create(AOwner:TComponent); override;
    procedure Init; override;
    procedure Exit; override;
    procedure First; override;
    procedure Next; override;
    procedure Open;
    procedure Close;
    function EOF: Boolean; override;
  published
    property CloseDataSource: Boolean read FCloseDataSource write FCloseDataSource default False;
    property DataSource: TDataSource read FDataSource write FDataSource;
    property OpenDataSource: Boolean read FOpenDataSource write FOpenDataSource default True;
    property RangeBegin;
    property RangeEnd;
    property RangeEndCount;
    property OnCheckEOF;
    property OnClose: TNotifyEvent read FOnClose write FOnClose;
    property OnFirst;
    property OnNext;
    property OnOpen: TNotifyEvent read FOnOpen write FOnOpen;
  end;


implementation

type
  EDSError = class(Exception);

constructor TfrDBDataSet.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  FOpenDataSource := True;
end;

procedure TfrDBDataSet.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FDataSource) then
    FDataSource := nil;
end;

function TfrDBDataSet.DataSet: TDataSet;
begin
  if (FDataSource <> nil) and (FDataSource.DataSet <> nil) then
    Result := FDataSource.DataSet
  else
  begin
    raise EDSError.Create('  Dataset');
    Result := nil;
  end;
end;

procedure TfrDBDataSet.Init;
begin
  Open;
  FBookmark := DataSet.GetBookmark;
  FEof := False;
end;

procedure TfrDBDataSet.Exit;
begin
  DataSet.GotoBookmark(FBookmark);
  DataSet.FreeBookmark(FBookmark);
  Close;
end;

procedure TfrDBDataSet.First;
begin
  if FRangeBegin = rbFirst then
    DataSet.First
  else if FRangeBegin = rbCurrent then
    DataSet.GotoBookmark(FBookmark);
  FEof := False;
  inherited First;
end;

procedure TfrDBDataSet.Next;
var b: TBookmark;
begin
  FEof := False;
  if FRangeEnd = reCurrent then
  begin
    b := DataSet.GetBookmark;
    if b = FBookmark then FEof := True;
    DataSet.FreeBookmark(b);
    Exit;
  end;
  DataSet.Next;
  inherited Next;
end;

function TfrDBDataSet.EOF: Boolean;
begin
  Result := inherited EOF or DataSet.EOF or FEof;
end;

procedure TfrDBDataSet.Open;
begin
  if FOpenDataSource then DataSet.Open;
  if Assigned(FOnOpen) then FOnOpen(Self);
end;

procedure TfrDBDataSet.Close;
begin
  if Assigned(FOnClose) then FOnClose(Self);
  if FCloseDataSource then DataSet.Close;
end;



end.
