
{*****************************************}
{                                         }
{             FastReport v2.2             }
{        Framed memo Add-In Object        }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_FMemo;

interface

{$I FR_Vers.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  FR_Class, StdCtrls, ExtCtrls, ClipBrd {$IFDEF Delphi4}, ImgList {$ENDIF},
  FR_Utils;

type
  TfrFramedMemoObject = class(TComponent)  // fake component
  end;

  TfrFrameLine = packed record
    Typ: Byte;
    Width: Byte;
  end;

  TfrFramedMemoView = class(TfrMemoView)
  public
    Frame: Array[0..3] of TfrFrameLine;
    constructor Create; override;
    procedure Assign(From:TfrView); override;
    procedure LoadFromStream(Stream:TStream); override;
    procedure SaveToStream(Stream:TStream); override;
    procedure CalcGaps; override;
    procedure ShowFrame; override;
  end;

  TFramedMemoForm = class(TfrObjEditorForm)
    GroupBox1: TGroupBox;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    ImageList1: TImageList;
    ComboBox1: TComboBox;
    ComboBox2: TComboBox;
    ComboBox3: TComboBox;
    ComboBox4: TComboBox;
    M1: TMemo;
    Button3: TButton;
    Button4: TButton;
    Button5: TButton;
    Button6: TButton;
    Image1: TImage;
    procedure ComboBox1DrawItem(Control: TWinControl; Index: Integer;
      Rect: TRect; State: TOwnerDrawState);
    procedure Button5Click(Sender: TObject);
    procedure M1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Button6Click(Sender: TObject);
    procedure FormActivate(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure ShowEditor(t:TfrView); override;
  end;


implementation

{$R *.DFM}

uses FR_Var, FR_Flds;

var
  FramedMemoForm: TFramedMemoForm;

{----------------------------------------------------------------------------}
constructor TfrFramedMemoView.Create;
var i: Integer;
begin
  inherited Create;
  Typ := gtAddIn;
  for i := 0 to 3 do
    Frame[i].Width := 1;
end;

procedure TfrFramedMemoView.Assign(From:TfrView);
begin
  inherited Assign(From);
  Frame := (From as TfrFramedMemoView).Frame;
end;

procedure TfrFramedMemoView.LoadFromStream(Stream:TStream);
begin
  inherited LoadFromStream(Stream);
  Stream.Read(Frame,SizeOf(Frame));
end;

procedure TfrFramedMemoView.SaveToStream(Stream:TStream);
begin
  inherited SaveToStream(Stream);
  Stream.Write(Frame,SizeOf(Frame));
end;

procedure TfrFramedMemoView.CalcGaps;
var i: Integer;
begin
  FrameWidth := 1;
  for i := 0 to 3 do
  with Frame[i] do
    if Width+Width*Typ*2 > FrameWidth then FrameWidth := Width+Width*Typ*2;
  inherited CalcGaps;
end;

procedure TfrFramedMemoView.ShowFrame;
procedure Line(x,y,dx,dy:Integer);
begin
  Canvas.MoveTo(x,y);
  Canvas.LineTo(x+dx,y+dy);
end;
procedure FrameLine(i: Integer);
begin
  Canvas.Pen.Width := Frame[i].Width;
  case i of
    0: if Frame[i].Typ = 0 then Line(x+dx,y,0,dy)
       else if Frame[i].Typ = 1 then
       begin
         Line(x+dx-Frame[i].Width,y,0,dy);
         Line(x+dx+Frame[i].Width,y,0,dy);
       end;
    1: if Frame[i].Typ = 0 then Line(x,y,0,dy)
       else if Frame[i].Typ = 1 then
       begin
         Line(x-Frame[i].Width,y,0,dy);
         Line(x+Frame[i].Width,y,0,dy);
       end;
    2: if Frame[i].Typ = 0 then Line(x,y+dy,dx,0)
       else if Frame[i].Typ = 1 then
       begin
         Line(x,y+dy-Frame[i].Width,dx,0);
         Line(x,y+dy+Frame[i].Width,dx,0);
       end;
    3: if Frame[i].Typ = 0 then Line(x,y,dx,0)
       else if Frame[i].Typ = 1 then
       begin
         Line(x,y-Frame[i].Width,dx,0);
         Line(x,y+Frame[i].Width,dx,0);
       end;
  end;
end;
begin
  if DisableDrawing then Exit;
  if (DocMode = dmPrinting) and ((FrameTyp and $F) = 0) then Exit;
  with Canvas do
  begin
    Pen.Style := psSolid;
    Brush.Style := bsClear;
    if DocMode = dmDesigning then
    begin
      Pen.Width := FrameWidth;
      Pen.Color := FillColor;
      Rectangle(x,y,x+dx+1,y+dy+1);
    end;
    Pen.Color := FrameColor;
    Pen.Width := Frame[0].Width;
    if ((FrameTyp and $F) = 0) and (DocMode = dmDesigning) then
    begin
      Pen.Color := clBlack;
      Pen.Width := 1;
      Line(x,y+3,0,-3); Line(x,y,4,0);
      Line(x,y+dy-4,0,3); Line(x,y+dy-1,4,0);
      Line(x+dx-4,y,3,0); Line(x+dx-1,y,0,4);
      Line(x+dx-4,y+dy-1,3,0); Line(x+dx-1,y+dy-1,0,-4);
    end
    else
    begin
      if ((FrameTyp and $F) = $F) and
        (Frame[0].Typ+Frame[1].Typ+Frame[2].Typ+Frame[3].Typ = 0) and
        (Frame[1].Width = Frame[0].Width) and
        (Frame[2].Width = Frame[0].Width) and
        (Frame[3].Width = Frame[0].Width) then
        Rectangle(x,y,x+dx+1,y+dy+1)
      else
      begin
        if (FrameTyp and $1) <> 0 then FrameLine(0);
        if (FrameTyp and $4) <> 0 then FrameLine(1);
        if (FrameTyp and $2) <> 0 then FrameLine(2);
        if (FrameTyp and $8) <> 0 then FrameLine(3);
      end;
    end;
  end;
end;


{-------------------------------------------------------------------------}
procedure TFramedMemoForm.ShowEditor(t:TfrView);
var c: Array[0..3] of TComboBox;
    i: Integer;
begin
  c[0] := ComboBox1; c[1] := ComboBox2;
  c[2] := ComboBox3; c[3] := ComboBox4;
  M1.Lines.Assign(t.Memo);
  with t as TfrFramedMemoView do
  begin
    for i := 0 to 3 do
      c[i].ItemIndex := Frame[i].Typ*3+Frame[i].Width-1;
    if ShowModal = mrOk then
    begin
      for i := 0 to 3 do
      begin
        Frame[i].Typ := c[i].ItemIndex div 3;
        Frame[i].Width := c[i].ItemIndex mod 3 + 1;
      end;
      Memo.Assign(M1.Lines);
    end;
  end;
end;

procedure TFramedMemoForm.FormActivate(Sender: TObject);
begin
  M1.SetFocus;
end;

procedure TFramedMemoForm.ComboBox1DrawItem(Control: TWinControl; Index: Integer;
  Rect: TRect; State: TOwnerDrawState);
begin
  with Control as TComboBox do
  begin
    Canvas.FillRect(Rect);
    ImageList1.BkColor := clWhite;
    if odSelected in State then
      ImageList1.ImageType := itMask else
      ImageList1.ImageType := itImage;
    ImageList1.Draw(Canvas,Rect.Left,Rect.Top,Index);
  end;
end;

procedure TFramedMemoForm.Button5Click(Sender: TObject);
begin
  VarForm := TVarForm.Create(nil);
  with VarForm do
  if ShowModal = mrOk then
  begin
    ClipBoard.Clear;
    if Variables.ItemIndex >= 0 then
      ClipBoard.AsText := '[' + Variables.Items[Variables.ItemIndex] + ']';
    M1.PasteFromClipboard;
  end;
  VarForm.Free;
  M1.SetFocus;
end;

procedure TFramedMemoForm.M1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = vk_Insert) and (Shift = []) then Button5Click(Self);
  if Key = vk_Escape then ModalResult := mrCancel;
end;

procedure TFramedMemoForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if (Key = vk_Return) and (ssCtrl in Shift) then
  begin
    ModalResult := mrOk;
    Key := 0;
  end;
end;

procedure TFramedMemoForm.Button6Click(Sender: TObject);
begin
  FieldsForm := TFieldsForm.Create(nil);
  with FieldsForm do
  if ShowModal = mrOk then
  begin
    ClipBoard.Clear;
    ClipBoard.AsText := '[' + DBField + ']';
    M1.PasteFromClipboard;
  end;
  FieldsForm.Free;
  M1.SetFocus;
end;

initialization
  FramedMemoForm := TFramedMemoForm.Create(nil);
  frRegisterObject(TfrFramedMemoView,FramedMemoForm.Image1.Picture.Bitmap,
    frLoadStr('SInsFMemo'),FramedMemoForm);

finalization
  FramedMemoForm.Free;

end.
