
{*****************************************}
{                                         }
{             FastReport v2.2             }
{             Object Inspector            }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_insp;

interface

{$I FR_Vers.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Buttons, FR_Class, FR_Ctrls, FR_Utils;

type
  TModifyEvent = procedure(Item:Integer) of object;

  TCtrlStyle = (csEdit,csDefEditor);

  TPropEditor = class(TSForm)
  public
    Str: String;
    View: TfrView;
    function ShowEditor:TModalResult; virtual;
  end;

  TProp = class
    Addr: PChar;
    Style: TCtrlStyle;
    Editor: TPropEditor;
    Enabled: Boolean;
    constructor Create(a:PChar; st:TCtrlStyle; de:TPropEditor); virtual;
  end;

  TInspForm = class(TSForm)
    Panel1: TPanel;
    PaintBox1: TPaintBox;
    SpeedButton1: TfrSpeedButton;
    Edit1: TEdit;
    procedure PaintBox1Paint(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure Edit1KeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure SpeedButton1Click(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormDeactivate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    FItems: TStringList;
    FItemIndex: Integer;
    FOnModify: TModifyEvent;
    w,w1: Integer;
    b: TBitmap;
    procedure SetItems(Value:TStringList);
    procedure SetItemIndex(Value:Integer);
    function GetCount:Integer;
    procedure DrawOneLine(i:Integer; a:Boolean);
    procedure SetItemValue(Value:String);
    function GetItemValue(i:Integer):String;
    function CurItem: TProp;
  public
    { Public declarations }
    V: TfrView;
    HideProperties: Boolean;
    procedure ClearItems;
    procedure ItemsChanged;
    procedure EnableItem(Index:Integer; Enable:Boolean);
    property Items: TStringList read FItems write SetItems;
    property ItemIndex: Integer read FItemIndex write SetItemIndex;
    property Count: Integer read GetCount;
    property OnModify: TModifyEvent read FOnModify write FOnModify;
  end;

var
  InspForm: TInspForm;

implementation

{$R *.DFM}

constructor TProp.Create(a:PChar; st:TCtrlStyle; de:TPropEditor);
begin
  inherited Create;
  Addr := a;
  Style := st;
  Editor := de;
  Enabled := True;
end;

function TPropEditor.ShowEditor:TModalResult;
begin
  Result := ShowModal;
end;

function TInspForm.CurItem: TProp;
begin
  Result := nil;
  if (FItemIndex <> -1) and (Count > 0) then
    Result := TProp(FItems.Objects[FItemIndex]);
end;

procedure TInspForm.SetItems(Value:TStringList);
begin
  FItems.Assign(Value);
  FItemIndex := -1;
  PaintBox1.Repaint;
  ItemIndex := 0;
end;

procedure TInspForm.SetItemValue(Value:String);
var p: TProp;
    s: String[255];
begin
  if HideProperties then Exit;
  p := TProp(FItems.Objects[FItemIndex]);
  s := Value;
  Move(s[0],p.Addr^,Ord(s[0])+1);
  if Assigned(FOnModify) then FOnModify(FItemIndex);
end;

function TInspForm.GetItemValue(i:Integer):String;
var p: TProp;
    s: String[255];
begin
  Result := '';
  p := TProp(FItems.Objects[i]);
  if p = nil then Exit;
  Move(p.Addr^,s[0],Ord(p.Addr^)+1);
  Result := s;
end;

procedure TInspForm.SetItemIndex(Value:Integer);
var ww: Integer;
begin
  if not TProp(FItems.Objects[Value]).Enabled then Exit;
  Edit1.Visible := (Count > 0) and not HideProperties;
  if (Count = 0) or (FItemIndex = Value) then Exit;
  if FItemIndex <> -1 then
    if Edit1.Modified then SetItemValue(Edit1.Text);
  if Value > Count-1 then Value := Count-1;
  FItemIndex := Value;
  SpeedButton1.Visible := (CurItem.Style = csDefEditor) and not HideProperties;
  Edit1.ReadOnly := CurItem.Style = csDefEditor;
  ww := w-w1-4;
  if SpeedButton1.Visible then
  begin
    SpeedButton1.SetBounds(w-16,2+FItemIndex*16+1,14,14);
    Dec(ww,15);
    Edit1.Text := '(' + FItems[FItemIndex] + ')';
  end
  else
    Edit1.Text := GetItemValue(FItemIndex);
  Edit1.SetBounds(w1+2,2+FItemIndex*16+1,ww,14);
  Edit1.SelectAll;
  Edit1.Modified := False;
  PaintBox1Paint(nil);
end;

function TInspForm.GetCount:Integer;
begin
  Result := FItems.Count;
end;

procedure TInspForm.ItemsChanged;
begin
  FItemIndex := -1;
  ItemIndex := 0;
end;

procedure TInspForm.EnableItem(Index:Integer; Enable:Boolean);
begin
  TProp(FItems.Objects[Index]).Enabled := Enable;
  PaintBox1Paint(nil);
end;

procedure TInspForm.DrawOneLine(i:Integer; a:Boolean);
procedure Line(x,y,dx,dy:Integer);
begin
  b.Canvas.MoveTo(x,y);
  b.Canvas.LineTo(x+dx,y+dy);
end;
begin
  if not TProp(FItems.Objects[i]).Enabled then Exit;
  if Count > 0 then
  with b.Canvas do
  begin
    Brush.Color := clBtnFace;
    Pen.Color := clBtnShadow;
    Font.Name := 'MS Sans Serif';
    Font.Size := 8;
    Font.Style := [];
    Font.Color := clBlack;
    if a then
    begin
      Pen.Color := clBtnShadow;
      Line(2,0+i*16,w-4,0);
      Line(w1-1,2+i*16,0,16);
      Pen.Color := clBlack;
      Line(2,1+i*16,w-4,0);
      Line(2,1+i*16,0,17);
      Pen.Color := clBtnHighlight;
      Line(3,17+i*16,w-5,0);
      Line(Edit1.Left,2+i*16,Edit1.Width,0);
      Line(w1,2+i*16,0,16);
      Line(w1+1,2+i*16,0,16);
      TextOut(7,3+i*16,FItems[i]);
    end
    else
    begin
      Line(2,17+i*16,w-4,0);
      Line(w1-1,2+i*16,0,16);
      Pen.Color := clBtnHighlight;
      Line(w1,2+i*16,0,16);
      TextOut(7,3+i*16,FItems[i]);
      Font.Color := clNavy;
      if TProp(FItems.Objects[i]).Style = csEdit then
        TextOut(w1+2,3+i*16,GetItemValue(i)) else
        TextOut(w1+2,3+i*16,'(' + FItems[i] + ')');
    end;
  end;
end;

procedure TInspForm.PaintBox1Paint(Sender: TObject);
var i: Integer;
    r: TRect;
begin
  r := PaintBox1.BoundsRect;
  b.Canvas.Brush.Color := clBtnFace;
  b.Canvas.FillRect(r);
  if not HideProperties then
  begin
    for i := 0 to Count-1 do
      if i <> FItemIndex then
        DrawOneLine(i,False);
    if FItemIndex <> -1 then DrawOneLine(FItemIndex,True);
  end;
  DrawEdge(b.Canvas.Handle,r,EDGE_SUNKEN,BF_RECT);
  PaintBox1.Canvas.Draw(0,0,b);
end;

procedure TInspForm.FormCreate(Sender: TObject);
begin
  w := PaintBox1.Width;
  w1 := w div 2;
  b := TBitmap.Create;
  b.Width := w;
  b.Height := PaintBox1.Height;
  SpeedButton1.Visible := False;
  FItemIndex := -1;
  FItems := TStringList.Create;
end;

procedure TInspForm.FormDestroy(Sender: TObject);
begin
  b.Free;
  ClearItems;
  FItems.Free;
end;

procedure TInspForm.ClearItems;
var i: Integer;
begin
  for i := 0 to FItems.Count-1 do
    TProp(FItems.Objects[i]).Free;
  FItems.Clear;
end;

procedure TInspForm.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if HideProperties then Exit;
  ItemIndex := y div 16;
  Edit1.SetFocus;
end;

procedure TInspForm.Edit1KeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if HideProperties then Exit;
  if Key = vk_Up then
  begin
    if ItemIndex > 0 then
      ItemIndex := ItemIndex-1;
    Key := 0;
  end
  else if Key = vk_Down then
  begin
    if ItemIndex < Count-1 then
      ItemIndex := ItemIndex+1;
    Key := 0;
  end
  else if Key = vk_Return then
  begin
    if CurItem.Style = csEdit then
    begin
      if Edit1.Modified then SetItemValue(Edit1.Text);
      Edit1.Modified := False;
    end
    else
      SpeedButton1Click(nil);
    Edit1.SelectAll;
    Key := 0;
  end;
end;

procedure TInspForm.SpeedButton1Click(Sender: TObject);
begin
  if HideProperties then Exit;
  with CurItem.Editor do
  begin
    Str := GetItemValue(FItemIndex);
    View := V;
    if ShowEditor = mrOk then
      SetItemValue(Str);
  end;
end;

procedure TInspForm.FormShow(Sender: TObject);
begin
  SetWindowPos(Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or
    SWP_NOSIZE or SWP_NOACTIVATE);
end;

procedure TInspForm.FormDeactivate(Sender: TObject);
begin
  if CurItem.Style = csEdit then
  begin
    if Edit1.Modified then SetItemValue(Edit1.Text);
    Edit1.Modified := False;
  end;
end;

end.
