
{*****************************************}
{                                         }
{             FastReport v2.2             }
{            OLE Add-In Object            }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_OLE;

interface

{$I FR_Vers.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OleCtnrs, StdCtrls, ExtCtrls, DB, FR_Class, FR_Utils,
{$IFDEF Delphi2}
  DBTables, Ole2;
{$ELSE}
  ActiveX;
{$ENDIF}


type
  TfrOLEObject = class(TComponent)  // fake component
  end;

  TfrOLEView = class(TfrView)
  public
    OleContainer: TOleContainer;
    constructor Create; override;
    procedure Assign(From:TfrView); override;
    destructor Destroy; override;
    procedure Draw(Canvas:TCanvas); override;
    procedure LoadFromStream(Stream:TStream); override;
    procedure SaveToStream(Stream:TStream); override;
    procedure Resized; override;
    procedure GetBlob(b:TField); override;
  end;

  TOleForm = class(TfrObjEditorForm)
    Button1: TButton;
    Button2: TButton;
    Button4: TButton;
    OleContainer1: TOleContainer;
    Image1: TImage;
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure ShowEditor(t:TfrView); override;
  end;

var
  OleForm: TOleForm;

implementation

{$R *.DFM}

procedure AssignOle(Cont1,Cont2: TOleContainer);
var st: TMemoryStream;
begin
  if Cont2.OleObjectInterface = nil then
  begin
    Cont1.DestroyObject;
    Exit;
  end;
  st := TMemoryStream.Create;
  Cont2.SaveToStream(st);
  st.Position := 0;
  Cont1.LoadFromStream(st);
  st.Free;
end;


{----------------------------------------------------------------------------}
constructor TfrOLEView.Create;
begin
  inherited Create;
  OleContainer := TOleContainer.Create(nil);
  with OleContainer do
  begin
    Parent := OleForm;
    Visible := False;
    AllowInPlace := False;
    AutoVerbMenu := False;
    BorderStyle := bsNone;
    SizeMode := smClip;
  end;
  Flags := 1;
end;

procedure TfrOLEView.Assign(From:TfrView);
begin
  inherited Assign(From);
  AssignOle(OleContainer,(From as TfrOLEView).OleContainer);
end;

destructor TfrOLEView.Destroy;
begin
  if OleForm <> nil then OleContainer.Free;
  inherited Destroy;
end;

procedure TfrOLEView.Draw(Canvas:TCanvas);
begin
  BeginDraw(Canvas);
  OleContainer.Width := dx;
  OleContainer.Height := dy;
  if DocMode = dmPrinting then
  begin
    Memo1.Assign(Memo);
    CurReport.InternalOnEnterRect(Memo1,Self);
  end;
  CalcGaps;
  with Canvas do
  begin
    ShowBackground;
    with OleContainer do
    if OleObjectInterface <> nil then
      OleDraw(OleObjectInterface, DVASPECT_CONTENT, Canvas.Handle, DRect)
    else if DocMode = dmDesigning then
    begin
      Font.Name := 'Arial';
      Font.Size := 8;
      Font.Style := [];
      Font.Color := clBlack;
      TextOut(x+2,y+2,'[OLE]');
    end;
    ShowFrame;
  end;
  if DocMode = dmPrinting then CurReport.InternalOnExportData(Self);
end;

procedure TfrOLEView.LoadFromStream(Stream:TStream);
var b: Byte;
begin
  inherited LoadFromStream(Stream);
  Stream.Read(b,1);
  if b <> 0 then
    OleContainer.LoadFromStream(Stream);
end;

procedure TfrOLEView.SaveToStream(Stream:TStream);
var b: Byte;
begin
  inherited SaveToStream(Stream);
  b := 0;
  if OleContainer.OleObjectInterface <> nil then
  begin
    b := 1;
    Stream.Write(b,1);
    OleContainer.SaveToStream(Stream);
  end
  else
    Stream.Write(b,1);
end;

procedure TfrOLEView.Resized;
var VS: TPoint;
begin
  if (Flags and flStretched) = 0 then
    with OleContainer do
    if OleObjectInterface <> nil then
    begin
      Run;
      VS.X := MulDiv(dx, 2540, Screen.PixelsPerInch);
      VS.Y := MulDiv(dy, 2540, Screen.PixelsPerInch);
      OleObjectInterface.SetExtent(DVASPECT_CONTENT, VS);
    end;
end;

procedure TfrOLEView.GetBlob(b:TField);
var s: TMemoryStream;
begin
  s := TMemoryStream.Create;
  (b as TBlobField).SaveToStream(s);
  s.Position := 0;
  OleContainer.LoadFromStream(s);
  s.Free;
end;


{----------------------------------------------------------------------------}
procedure TOleForm.ShowEditor(t:TfrView);
begin
  AssignOle(OleContainer1, (t as TfrOLEView).OleContainer);
  ShowModal;
  AssignOle((t as TfrOLEView).OleContainer, OleContainer1);
  OleContainer1.DestroyObject;
end;

procedure TOleForm.Button1Click(Sender: TObject);
begin
  with OleContainer1 do
    if InsertObjectDialog then
      DoVerb(PrimaryVerb);
end;

procedure TOleForm.Button2Click(Sender: TObject);
begin
  OleContainer1.DoVerb(ovPrimary);
end;


initialization
  OleForm := TOleForm.Create(nil);
  frRegisterObject(TfrOLEView,OleForm.Image1.Picture.Bitmap,
    frLoadStr('SInsOLEObject'),OleForm);

finalization
  OleForm.Free;
  OleForm := nil;

end.
