
{*****************************************}
{                                         }
{             FastReport v2.2             }
{            Various routines             }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_Utils;

interface

{$I FR_Vers.inc}

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls,
  DB, Forms, StdCtrls, ClipBrd, IniFiles, Menus;

type
  TSForm = class(TForm)
  private
//    procedure SaveLanguage;
    procedure LoadLanguage;
  public
    constructor Create(AOwner: TComponent); override;
    procedure AfterLanguageLoaded; virtual;
  end;


procedure frReadMemo(Stream:TStream; l:TStrings);
procedure frWriteMemo(Stream:TStream; l:TStrings);
procedure frEnableControls(c: Array of TControl; e:Boolean);
function frGetDataSet(ComplexName:String): TDataSet;
procedure frGetDataSetAndField(ComplexName:String; var DataSet:TDataSet;
  var Field:TField);
function frLoadStr(sName:String): String;
function frGetWindowsVersion: String;

var
  Ini: TIniFile;


implementation

uses FR_Class;

type
  THackControl = class(TControl)
  end;

procedure CreateIniFile;
begin
  Ini := TIniFile.Create('FR.lng');
end;

procedure TSForm.LoadLanguage;
var i: Integer;
    c: TComponent;
    cFullName,cPropName,cName: String;
    sl: TStringList;
begin
  if Ini = nil then CreateIniFile;
  sl := TStringList.Create;
  Caption := Ini.ReadString(Name, 'Caption', Caption);
  Ini.ReadSection(Name, sl);
  for i := 0 to sl.Count-1 do
  begin
    cFullName := AnsiUpperCase(sl[i]);
    if cFullName <> 'CAPTION' then
    begin
      cName := Copy(cFullName, 1, Pos('.',cFullName)-1);
      cPropName := Copy(cFullName, Pos('.',cFullName)+1, 255);
      c := FindComponent(cName);
      if c <> nil then
      begin
        if c is TMenuItem then
        begin
          if cPropName = 'CAPTION' then
            TMenuItem(c).Caption := Ini.ReadString(Name, cFullName, TMenuItem(c).Caption)
        end
        else
          if cPropName = 'CAPTION' then
            THackControl(c).Caption := Ini.ReadString(Name, cFullName, THackControl(c).Caption)
          else if cPropName = 'HINT' then
            THackControl(c).Hint := Ini.ReadString(Name, cFullName, THackControl(c).Hint);
      end;
    end;
  end;
  sl.Free;
  AfterLanguageLoaded;
end;

{procedure TSForm.SaveLanguage;
var i: Integer;
    c: THackControl;
begin
  if Ini = nil then CreateIniFile;
  Ini.WriteString(Name, 'Caption', Caption);
  for i := 0 to ComponentCount-1 do
  begin
    if Components[i] is TControl then
    begin
      c := THackControl(Components[i]);
      if c.Caption <> '' then
        Ini.WriteString(Name, c.Name + '.Caption', c.Caption);
      if c.Hint <> '' then
        Ini.WriteString(Name, c.Name + '.Hint', c.Hint);
    end
    else if Components[i] is TMenuItem then
    with Components[i] as TMenuItem do
      if Caption <> '-' then
        Ini.WriteString(Self.Name, Name + '.Caption', Caption);
  end;
end;}

constructor TSForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  LoadLanguage;
end;

procedure TSForm.AfterLanguageLoaded;
begin
// abstract method
end;


function frLoadStr(sName:String): String;
begin
  if Ini = nil then CreateIniFile;
  Result := Ini.ReadString('Strings', sName, sName);
end;

procedure RemoveQuotes(var s:String);
begin
  if (s[1] = '"') and (s[Length(s)] = '"') then
    s := Copy(s,2,Length(s)-2);
end;

procedure frReadMemo(Stream:TStream; l:TStrings);
var s: String[255];
    i: Integer;
    b: Byte;
begin
  l.Clear;
  i := 1;
  repeat
    Stream.Read(b,1);
    if (b = 13) or (b = 0) then
    begin
      s[0] := Chr(i-1);
      if not ((b = 0) and (i = 1)) then l.Add(s);
      i := 1;
    end
    else if b <> 0 then
    begin
      s[i] := Chr(b);
      Inc(i);
    end;
  until b = 0;
end;

procedure frWriteMemo(Stream:TStream; l:TStrings);
var s: String[255];
    i,j: Integer;
    b: Byte;
begin
  for i := 0 to l.Count-1 do
  begin
    s := l[i];
    for j := 1 to Length(s) do
      Stream.Write(s[j],1);
    b := 13;
    if i <> l.Count-1 then Stream.Write(b,1);
  end;
  b := 0;
  Stream.Write(b,1);
end;

procedure frEnableControls(c: Array of TControl; e:Boolean);
const
  Clr1: Array[Boolean] of TColor = (clGrayText,clWindowText);
  Clr2: Array[Boolean] of TColor = (clBtnFace,clWindow);
var i: Integer;
begin
  for i := Low(c) to High(c) do
    if c[i] is TLabel then
      with c[i] as TLabel do
      begin
        Font.Color := Clr1[e];
        Enabled := e;
      end
    else if c[i] is TEdit then
      with c[i] as TEdit do
      begin
        Color := Clr2[e];
        Enabled := e;
      end;
end;

function frGetDataSet(ComplexName:String): TDataSet;
var n: Integer;
    f: TComponent;
    s1,s2: String;
begin
  Result := nil;
  f := CurReport.Owner;
  n := Pos('.',ComplexName);
  try
    if n = 0 then
      Result := f.FindComponent(ComplexName) as TDataSet
    else
    begin
      s1 := Copy(ComplexName, 1, n-1);        // module name
      s2 := Copy(ComplexName, n+1, 255);      // database name
      f := FindGlobalComponent(s1);
      if f <> nil then
        Result := f.FindComponent(s2) as TDataSet;
    end;
  except
    on Exception do
      raise EClassNotFound.Create(' ' + ComplexName);
  end;
end;

procedure frGetDataSetAndField(ComplexName:String; var DataSet:TDataSet;
  var Field:TField);
var n: Integer;
    f: TComponent;
    s1,s2,s3: String;
begin
  Field := nil;
  f := CurReport.Owner;
  n := Pos('.',ComplexName);
  if n <> 0 then
  begin
    s1 := Copy(ComplexName, 1, n-1);        // table name
    s2 := Copy(ComplexName, n+1, 255);      // field name
    if Pos('.',s2) <> 0 then                // module name present
    begin
      s3 := Copy(s2,Pos('.',s2)+1,255);
      s2 := Copy(s2,1,Pos('.',s2)-1);
      f := FindGlobalComponent(s1);
      if f <> nil then
      begin
        DataSet := f.FindComponent(s2) as TDataSet;
        RemoveQuotes(s3);
        if DataSet <> nil then
          Field := DataSet.FieldByName(s3);
      end;
    end
    else
    begin
      DataSet := f.FindComponent(s1) as TDataSet;
      RemoveQuotes(s2);
      if DataSet <> nil then
        Field := DataSet.FieldByName(s2);
    end;
  end
  else if DataSet <> nil then
  begin
    RemoveQuotes(ComplexName);
    Field := DataSet.FindField(ComplexName);
  end;
end;

function frGetWindowsVersion: String;
var Ver: TOsVersionInfo;
begin
  Ver.dwOSVersionInfoSize := SizeOf(Ver);
  GetVersionEx(Ver);
  with Ver do begin
    case dwPlatformId of
      VER_PLATFORM_WIN32s: Result := '32s';
      VER_PLATFORM_WIN32_WINDOWS:
        begin
          dwBuildNumber := dwBuildNumber and $0000FFFF;
          if (dwMajorVersion > 4) or ((dwMajorVersion = 4) and
            (dwMinorVersion >= 10)) then
            Result := '98' else
            Result := '95';
        end;
      VER_PLATFORM_WIN32_NT: Result := 'NT';
    end;
  end;
end;


initialization

finalization
  if Ini <> nil then Ini.Free;

end.
