
{*****************************************}
{                                         }
{             FastReport v2.2             }
{             Report preview              }
{                                         }
{  Copyright (c) 1998-99 by Tzyganenko A. }
{                                         }
{*****************************************}

unit FR_View;

interface

{$I FR_Vers.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, Buttons, StdCtrls, Menus, FR_Class, FR_Ctrls, FR_Utils;

type
  TPreviewForm = class(TSForm)
    TPanel: TPanel;
    ExitBtn: TfrSpeedButton;
    ProcBtn: TfrSpeedButton;
    ProcMenu: TPopupMenu;
    N2001: TMenuItem;
    N1501: TMenuItem;
    N1001: TMenuItem;
    N751: TMenuItem;
    N501: TMenuItem;
    N251: TMenuItem;
    N101: TMenuItem;
    N1: TMenuItem;
    N2: TMenuItem;
    N3: TMenuItem;
    LoadBtn: TfrSpeedButton;
    SaveBtn: TfrSpeedButton;
    PrintBtn: TfrSpeedButton;
    OpenDialog: TOpenDialog;
    SaveDialog: TSaveDialog;
    PrintDialog: TPrintDialog;
    Bevel2: TBevel;
    ScrollBox1: TScrollBox;
    Panel1: TPanel;
    PBox: TPaintBox;
    RPanel: TPanel;
    PgUp: TfrSpeedButton;
    PgDown: TfrSpeedButton;
    VScrollBar: TScrollBar;
    BPanel: TPanel;
    Bevel1: TBevel;
    Label1: TLabel;
    HScrollBar: TScrollBar;
    FindBtn: TfrSpeedButton;
    procedure FormResize(Sender: TObject);
    procedure PBoxPaint(Sender: TObject);
    procedure VScrollBarChange(Sender: TObject);
    procedure HScrollBarChange(Sender: TObject);
    procedure PgUpClick(Sender: TObject);
    procedure PgDownClick(Sender: TObject);
    procedure ProcBtnClick(Sender: TObject);
    procedure N3Click(Sender: TObject);
    procedure ExitBtnClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure PBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure LoadBtnClick(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure PrintBtnClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FindBtnClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
  private
    { Private declarations }
    Doc: TfrReport;
    CurPage: Integer;
    procedure ShowPageNum;
    procedure SetToCurPage;
    procedure WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo); message WM_GETMINMAXINFO;
    procedure RedrawAll;
  public
    { Public declarations }
    procedure LoadFromFile(name:String);
    procedure SaveToFile(name:String);
    procedure Show_Modal(ADoc:TfrReport);
  end;

var
  PreviewForm: TPreviewForm;

implementation

{$R *.DFM}
uses Printers, FR_Prntr, FR_Srch;

var
  wx,wy,nx: Integer;
  ofx,ofy,OldV,OldH: Integer;
  rg: HRgn;
  per: Double;
  mode: (mdNone,mdPageWidth,mdOnePage,mdTwoPages);
  PaintAllowed: Boolean;
  PaintPart: Boolean;
  FindStr: String;
  StrFound: Boolean;
  StrBounds: TRect;
  DrawMode: (dmDraw,dmFind);


{----------------------------------------------------------------------------}
// fast routine for emf drawing
procedure DrawEMF(Canvas:TCanvas; emf:TMetafile; DrawRect:TRect);
  function EnumEMFRecordsProc(DC:HDC; HandleTable:PHandleTable;
    EMFRecord:PEnhMetaRecord; nObj:Integer; OptData:Pointer):Bool; stdcall;
  var CanDraw: Boolean;
      Typ: Byte;
      r: TRect;
      s: String;
      t: TEMRExtTextOut;
  begin
    Result := True;
    if DrawMode = dmFind then
      CanDraw := False else
      CanDraw := True;
    Typ := EMFRecord^.iType;
    if DrawMode = dmDraw then
      if Typ in [42..47,76..81,83,84] then
      begin
        r := PEMRExtTextOut(EMFRecord)^.rclBounds;
        InflateRect(r,1,1);
        CanDraw := RectVisible(DC,r);
      end;
    if DrawMode = dmFind then
      if Typ in [83,84] then
      begin
        t := PEMRExtTextOut(EMFRecord)^;
        s := WideCharLenToString(PWideChar(PChar(EMFRecord)+t.EMRText.offString),
          t.EMRText.nChars);
        if not PreviewSearchForm.CB1.Checked then s := AnsiUpperCase(s);
        StrFound := Pos(FindStr,s) <> 0;
        if StrFound then
        begin
          StrBounds := t.rclBounds;
          Result := False;
        end;
      end;
    if CanDraw then
      PlayEnhMetafileRecord(DC,HandleTable^,EMFRecord^,nObj);
  end;
begin
  EnumEnhMetafile(Canvas.Handle,emf.Handle,@EnumEMFRecordsProc,nil,DrawRect);
end;

procedure TPreviewForm.FormCreate(Sender: TObject);
begin
  PreviewSearchForm := TPreviewSearchForm.Create(nil);
  PrintBtn.Enabled := Printer.Printers.Count > 0;
end;

procedure TPreviewForm.FormDestroy(Sender: TObject);
begin
  PreviewSearchForm.Free;
end;

procedure TPreviewForm.Show_Modal(ADoc:TfrReport);
begin
  Doc := ADoc;
  RedrawAll;
  HScrollBar.Position := 0;
  VScrollBar.Position := 0;
  ShowModal;
end;

procedure TPreviewForm.WMGetMinMaxInfo(var Msg: TWMGetMinMaxInfo);
begin
  with Msg.MinMaxInfo^ do
  begin
    ptMaxSize.x := Screen.Width;
    ptMaxSize.y := Screen.Height;
    ptMaxPosition.x := 0;
    ptMaxPosition.y := 0;
  end;
end;

procedure TPreviewForm.RedrawAll;
begin
  per := 1;
  DrawMode := dmDraw;
  N1001.Checked := True;
  mode := mdNone;
  CurPage := 1;
  ShowPageNum;
  ofx := 0; ofy := 0; OldH := 0; OldV := 0;
  HScrollBar.Position := 0;
  VScrollBar.Position := 0;
  FormResize(nil);
  PaintPart := False;
  PBoxPaint(nil);
end;

procedure TPreviewForm.FormResize(Sender: TObject);
var i,j,y,d,dwx,dwy,maxx,maxy,maxdy,curx: Integer;
begin
  PaintAllowed := False;
  wx := PBox.Width; wy := PBox.Height;
  with Doc.EMFPages[CurPage-1].PrnInfo do
  begin
    dwx := Pgw; dwy := Pgh;
  end;
  case mode of
    mdNone:;
    mdPageWidth: per := (wx - 20)/dwx;
    mdOnePage: per := (wy - 20)/dwy;
    mdTwoPages: per := (wx - 30)/(2*dwx);
  end;
  ProcBtn.Caption := IntToStr(Round(per*100)) + '%';
  nx := 0; maxx := 10; j := 0;
  for i := 0 to Doc.EMFPages.Count-1 do
  begin
    d := maxx + 10 + Round(Doc.EMFPages[i].PrnInfo.Pgw*per);
    if d > wx then
    begin
      if nx < j then nx := j;
      j := 0; maxx := 10;
    end
    else
    begin
      maxx := d;
      Inc(j);
      if i = Doc.EMFPages.Count-1 then
        if nx < j then nx := j;
    end;
  end;
  if nx = 0 then nx := 1;
  if mode = mdOnePage then nx := 1;
  if mode = mdTwoPages then nx := 2;
  y := 10;
  i := 0;
  maxx := 0; maxy := 0;
  while i < Doc.EMFPages.Count do
  begin
    j := 0; maxdy := 0; curx := 10;
    while (j < nx) and (i+j < Doc.EMFPages.Count) do
    begin
      dwx := Round(Doc.EMFPages[i+j].PrnInfo.Pgw*per);
      dwy := Round(Doc.EMFPages[i+j].PrnInfo.Pgh*per);
      if (nx = 1) and (dwx < wx) then
      begin
        d := (wx - dwx) div 2;
        Doc.EMFPages[i+j].r := Rect(d,y,d+dwx,y+dwy);
      end
      else
        Doc.EMFPages[i+j].r := Rect(curx,y,curx+dwx,y+dwy);
      if maxx < Doc.EMFPages[i+j].r.Right then
        maxx := Doc.EMFPages[i+j].r.Right;
      if maxy < Doc.EMFPages[i+j].r.Bottom then
        maxy := Doc.EMFPages[i+j].r.Bottom;
      Inc(j);
      if maxdy < dwy then maxdy := dwy;
      Inc(curx,dwx+10);
    end;
    Inc(y,maxdy + 10);
    Inc(i,nx);
  end;
  PgDown.Top := RPanel.Height-16;
  PgUp.Top := PgDown.Top-16;
  VScrollBar.Height := PgUp.Top-1;
  HScrollBar.Width := BPanel.Width-HScrollBar.Left-VScrollBar.Width;
  maxx := maxx-wx;
  maxy := maxy-wy;
  if maxx < 0 then maxx := 0 else Inc(maxx,10);
  if maxy < 0 then maxy := 0 else Inc(maxy,10);
  HScrollBar.Max := maxx; VScrollBar.Max := maxy;
  HScrollBar.Enabled := maxx <> 0;
  VScrollBar.Enabled := maxy <> 0;
  SetToCurPage;
  PaintAllowed := True;
end;

procedure TPreviewForm.SetToCurPage;
begin
  if ofy <> Doc.EMFPages[CurPage-1].r.Top-10 then
    VScrollBar.Position := Doc.EMFPages[CurPage-1].r.Top-10;
end;

procedure TPreviewForm.ShowPageNum;
begin
  Label1.Caption := frLoadStr('SPg') + ' ' + IntToStr(CurPage) + '/' +
    IntToStr(Doc.EMFPages.Count);
end;

procedure TPreviewForm.PBoxPaint(Sender: TObject);
var i: Integer;
    r,r1: TRect;
begin
  if not PaintAllowed then Exit;
  if not PaintPart then
  begin
    rg := CreateRectRgn(0,0,wx,wy);
    SelectObject(PBox.Canvas.Handle,rg);
    with PBox.Canvas do
    begin
      Brush.Color := clGray;
      FillRect(Rect(0,0,wx,wy));
      Pen.Color := clBlack;
      Brush.Color := clWhite;
    end;
  end;
  GetRgnBox(rg,r);
  Inc(r.Left,ofx); Inc(r.Top,ofy);
  Inc(r.Right,ofx); Inc(r.Bottom,ofy);
  if PaintPart then
  begin
    for i := 0 to Doc.EMFPages.Count-1 do
      with Doc.EMFPages[i].r do
      if not ((Left >= r.Right) or (Top >= r.Bottom) or
              (Right <= r.Left) or (Bottom <= r.Top)) then
        ExcludeClipRect(PBox.Canvas.Handle,Left-ofx,Top-ofy,
                        Right-ofx+1,Bottom-ofy+1);
    with PBox.Canvas do
    begin
      Brush.Color := clGray;
      FillRect(Rect(0,0,wx,wy));
      Pen.Color := clBlack;
      Brush.Color := clWhite;
      SelectObject(Handle,rg);
    end;
  end;
  for i := 0 to Doc.EMFPages.Count-1 do
  begin
    with Doc.EMFPages[i].r do
    if not ((Left >= r.Right) or (Top >= r.Bottom) or
            (Right <= r.Left) or (Bottom <= r.Top)) then
    with PBox.Canvas do
    begin
      Rectangle(Left-ofx,Top-ofy,Right-ofx,Bottom-ofy);
      if Doc.EMFPages[i].pgMargins then
         DrawEMF(PBox.Canvas,Doc.EMFPages[i].Pict,
           Rect(Left-ofx+1,Top-ofy+1,Right-ofx-1,Bottom-ofy-1))
      else
      begin
        with Doc.EMFPages[i].PrnInfo do
        begin
          r1.Left := Round(Ofx*per);
          r1.Top := Round(Ofy*per);
          r1.Right := r1.Left + Round(Pw*per);
          r1.Bottom := r1.Top + Round(Ph*per);
        end;
        Inc(r1.Left,Left-ofx+1); Inc(r1.Right,Left-ofx+1);
        Inc(r1.Top,Top-ofy+1); Inc(r1.Bottom,Top-ofy+1);
        DrawEMF(PBox.Canvas,Doc.EMFPages[i].Pict,r1);
      end;
      Polyline([Point(Left-ofx+2,Bottom-ofy),Point(Right-ofx,Bottom-ofy),
                Point(Right-ofx,Top-ofy+2)]);
    end;
  end;
  if not PaintPart then DeleteObject(rg);
end;

procedure TPreviewForm.VScrollBarChange(Sender: TObject);
var i,p,pp: Integer;
begin
  p := VScrollBar.Position;
  ofy := p;
  if p < OldV then {scroll up}
  begin
    PaintPart := True;
    pp := OldV-p;
    if OldV-p > wy then pp := wy;
    BitBlt(PBox.Canvas.Handle,0,pp,wx,wy-pp,
           PBox.Canvas.Handle,0,0,srcCopy);
    rg := CreateRectRgn(0,0,wx,pp);
    SelectObject(PBox.Canvas.Handle,rg);
    PBoxPaint(nil);
    DeleteObject(rg);
    PaintPart := False;
  end
  else if p > OldV then {scroll down}
  begin
    PaintPart := True;
    pp := p-OldV;
    if p-OldV > wy then pp := wy;
    BitBlt(PBox.Canvas.Handle,0,0,wx,wy-pp,
           PBox.Canvas.Handle,0,pp,srcCopy);
    rg := CreateRectRgn(0,wy-pp,wx,wy+1);
    SelectObject(PBox.Canvas.Handle,rg);
    PBoxPaint(nil);
    DeleteObject(rg);
    PaintPart := False;
  end;
  OldV := p;
  for i := 0 to Doc.EMFPages.Count-1 do
    if (Doc.EMFPages[i].r.Top < ofy+11) and
      (Doc.EMFPages[i].r.Bottom > ofy+11) then
    begin
      CurPage := i+1;
      ShowPageNum;
      break;
    end;
end;

procedure TPreviewForm.HScrollBarChange(Sender: TObject);
var p: Integer;
begin
  p := HScrollBar.Position;
  ofx := p;
  if p < OldH then {scroll left}
  begin
    PaintPart := True;
    BitBlt(PBox.Canvas.Handle,(OldH-p),0,wx-(OldH-p),wy,
           PBox.Canvas.Handle,0,0,srcCopy);
    rg := CreateRectRgn(0,0,(OldH-p)+1,wy);
    SelectObject(PBox.Canvas.Handle,rg);
    PBoxPaint(nil);
    DeleteObject(rg);
    PaintPart := False;
  end
  else if p > OldH then {scroll right}
  begin
    PaintPart := True;
    BitBlt(PBox.Canvas.Handle,0,0,wx-(p-OldH),wy,
           PBox.Canvas.Handle,(p-OldH),0,srcCopy);
    rg := CreateRectRgn(wx-(p-OldH)-1,0,wx+1,wy);
    SelectObject(PBox.Canvas.Handle,rg);
    PBoxPaint(nil);
    DeleteObject(rg);
    PaintPart := False;
  end;
  OldH := p;
end;

procedure TPreviewForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key in [vk_Up,vk_Down,vk_Prior,vk_Next] then
    if VScrollBar.Enabled then VScrollBar.SetFocus;
  if Key in [vk_Left,vk_Right] then
    if HScrollBar.Enabled then HScrollBar.SetFocus;
  if Key = vk_Up then
    VScrollBar.Position := VScrollBar.Position-VScrollBar.SmallChange
  else if Key = vk_Down then
    VScrollBar.Position := VScrollBar.Position+VScrollBar.SmallChange
  else if Key = vk_Left then
    HScrollBar.Position := HScrollBar.Position-HScrollBar.SmallChange
  else if Key = vk_Right then
    HScrollBar.Position := HScrollBar.Position+HScrollBar.SmallChange
  else if Key = vk_Prior then
    if ssCtrl in Shift then
      PgUpClick(nil)
    else
      VScrollBar.Position := VScrollBar.Position-VScrollBar.LargeChange
  else if Key = vk_Next then
    if ssCtrl in Shift then
      PgDownClick(nil)
    else
      VScrollBar.Position := VScrollBar.Position+VScrollBar.LargeChange
  else if Key = vk_Space then
    ProcBtnClick(nil)
  else if Key = vk_Escape then
    ExitBtnClick(nil)
  else if ssCtrl in Shift then
  begin
    if Chr(Key) = 'L' then LoadBtnClick(nil)
    else if Chr(Key) = 'S' then SaveBtnClick(nil)
    else if Chr(Key) = 'P' then PrintBtnClick(nil)
    else if Chr(Key) = 'F' then FindBtnClick(nil)
  end
  else Exit;
  Key := 0;
end;

procedure TPreviewForm.PBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var i: Integer;
begin
  if Button in [mbLeft] then
  for i := 0 to Doc.EMFPages.Count-1 do
    if PtInRect(Doc.EMFPages[i].r,Point(x+ofx,y+ofy)) then
    begin
      CurPage := i+1;
      SetToCurPage;
      CurPage := i+1;
      ShowPageNum;
      break;
    end;
end;

procedure TPreviewForm.PgUpClick(Sender: TObject);
begin
  if CurPage > 1 then Dec(CurPage);
  ShowPageNum;
  SetToCurPage;
end;

procedure TPreviewForm.PgDownClick(Sender: TObject);
begin
  if CurPage < Doc.EMFPages.Count then Inc(CurPage);
  ShowPageNum;
  SetToCurPage;
end;

procedure TPreviewForm.ProcBtnClick(Sender: TObject);
var pt: TPoint;
begin
  pt := ClientToScreen(Point(ProcBtn.Left,ProcBtn.Top+ProcBtn.Height+2));
  ProcMenu.Popup(pt.x,pt.y);
end;

procedure TPreviewForm.N3Click(Sender: TObject);
begin
  ofx := 0;
  with Sender as TMenuItem do
  begin
    case Tag of
      1: mode := mdPageWidth;
      2: mode := mdOnePage;
      3: mode := mdTwoPages;
    else
      begin
        mode := mdNone;
        per := Tag/100;
      end;
    end;
    Checked := True;
  end;
  FormResize(nil);
  PBoxPaint(nil);
end;

procedure TPreviewForm.LoadBtnClick(Sender: TObject);
begin
  OpenDialog.Filter := frLoadStr('SRepFile') + ' (*.frp)|*.frp';
  with OpenDialog do
   if Execute then
     LoadFromFile(FileName);
end;

procedure TPreviewForm.SaveBtnClick(Sender: TObject);
var i: Integer;
    s: String;
begin
  s := frLoadStr('SRepFile') + ' (*.frp)|*.frp';
  for i := 0 to frFiltersCount-1 do
    s := s + '|' + frFilters[i].FilterDesc + '|' + frFilters[i].FilterExt;
  with SaveDialog do
  begin
    Filter := s;
    FilterIndex := 1;
    if Execute then
      if FilterIndex = 1 then
        SaveToFile(FileName)
      else
      if Doc.CanRebuild then
        begin
          Doc.ExportTo(frFilters[FilterIndex-2].ClassRef,
            ChangeFileExt(FileName,Copy(frFilters[FilterIndex-2].FilterExt,2,255)));
          RedrawAll;
        end;
  end;
end;

procedure TPreviewForm.PrintBtnClick(Sender: TObject);
var FromPg,ToPg: Integer;
    ind: Integer;
begin
  ind := Printer.PrinterIndex;
  with PrintDialog do
  begin
    Copies := 1;
    Collate := True;
    FromPage := 1;
    ToPage := Doc.EMFPages.Count;
    MaxPage := Doc.EMFPages.Count;
    PaintAllowed := True;
    if Execute then
    begin
      if Printer.PrinterIndex <> ind then
        if Doc.CanRebuild then
          if Doc.ChangePrinter(ind,Printer.PrinterIndex) then
          begin
            Doc.PrepareReport;
            RedrawAll;
          end
          else exit;
      FromPg := 1; ToPg := Doc.EMFPages.Count;
      if PrintDialog.PrintRange = prPageNums then
      begin
        FromPg := PrintDialog.FromPage;
        ToPg := PrintDialog.ToPage;
      end;
      Doc.PrintPreparedReport(FromPg,ToPg,1,True);
    end;
  end;
  PBoxPaint(nil);
end;

procedure TPreviewForm.ExitBtnClick(Sender: TObject);
begin
  ModalResult := mrOk;
end;

procedure TPreviewForm.LoadFromFile(name:String);
begin
  Doc.LoadPreparedReport(name);
  CurPage := 1;
  FormResize(nil);
  PaintAllowed := False;
  ShowPageNum;
  SetToCurPage;
  PaintAllowed := True;
  PBoxPaint(nil);
end;

procedure TPreviewForm.SaveToFile(name:String);
begin
  name := ChangeFileExt(name,'.frp');
  Doc.SavePreparedReport(name);
  FormResize(nil);
  PaintAllowed := True;
  PBoxPaint(nil);
end;

procedure TPreviewForm.FindBtnClick(Sender: TObject);
var i: Integer;
begin
  if PreviewSearchForm.ShowModal = mrOk then
  with PreviewSearchForm do
  begin
    FindStr := Edit1.Text;
    if not CB1.Checked then FindStr := AnsiUpperCase(FindStr);
    StrFound := False;
    PaintAllowed := False;
    DrawMode := dmFind;
    if RB1.Checked then
      i := 0 else
      i := CurPage;
    while i < Doc.EMFPages.Count do
    begin
      DrawEMF(PBox.Canvas,Doc.EMFPages[i].Pict,Rect(0,0,10,10));
      if StrFound then
      begin
        PaintAllowed := True;
        DrawMode := dmDraw;
        CurPage := i+1;
        ShowPageNum;
        VScrollBar.Position := Doc.EMFPages[i].r.Top + Round(StrBounds.Top*per) - 10;
        HScrollBar.Position := Doc.EMFPages[i].r.Left + Round(StrBounds.Left*per) - 10;
        break;
      end;
      Inc(i);
    end;
  end;
  PaintAllowed := True;
  DrawMode := dmDraw;
end;

end.
